from __future__ import annotations

from collections.abc import Callable, Mapping, Sequence
from typing import Any, ParamSpec, TypeVar, overload

from tabb.base import BaseCommand
from tabb.command import Command
from tabb.group import Group

T = TypeVar("T")
P = ParamSpec("P")


@overload
def command(callback: Callable[P, T], /) -> Command[P, T]:
    ...


@overload
def command(
    name: str | None = None,
    *,
    add_help_option: bool = True,
    deprecated: bool = False,
    epilog: str | None = None,
    help: str | None = None,
    hidden: bool = False,
    import_dependencies: str | None = None,
    options_metavar: str | None = "[OPTIONS]",
    package: str | None = None,
    short_help: str | None = None,
) -> Callable[[Callable[P, T]], Command[P, T]]:
    ...


def command(
    name: Callable[P, T] | str | None = None,
    **kwargs: Any,
) -> Command[P, T] | Callable[[Callable[P, T]], Command[P, T]]:
    if callable(name):
        return Command(name)

    def decorator(callback: Callable[P, T]) -> Command[P, T]:
        if isinstance(callback, Command):
            raise TypeError("Attempted to convert a callback into a command twice.")

        return Command(callback, name=name, **kwargs)

    return decorator


@overload
def group(callback: Callable[..., Any], /) -> Group[Any]:
    ...


@overload
def group(
    name: str | None = None,
    commands: Mapping[str, BaseCommand[T]] | Sequence[BaseCommand[T]] | None = None,
    *,
    add_help_option: bool = True,
    deprecated: bool = False,
    epilog: str | None = None,
    help: str | None = None,
    hidden: bool = False,
    import_dependencies: str | None = None,
    options_metavar: str | None = "[OPTIONS]",
    package: str | None = None,
    short_help: str | None = None,
) -> Callable[[Callable[..., Any]], Group[Any]]:
    ...


def group(
    name: Callable[..., Any] | str | None = None,
    **kwargs: Any,
) -> Group[Any] | Callable[[Callable[..., Any]], Group[Any]]:
    if callable(name):
        return Group(name)

    def decorator(callback: Callable[..., Any]) -> Group[Any]:
        if isinstance(callback, Group):
            raise TypeError("Attempted to convert a callback into a group twice.")

        return Group(callback, name=name, **kwargs)

    return decorator



from __future__ import annotations

import sys
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, TextIO

if TYPE_CHECKING:
    from tabb.context import Context
    from tabb.parameter import ParserParameter


@contextmanager
def augment_usage_errors(
    ctx: Context[Any],
    param: ParserParameter[Any] | None = None,
    *,
    parameter_errors: type[Exception] | tuple[type[Exception], ...] = (),
    usage_errors: type[Exception] | tuple[type[Exception], ...] = (),
) -> Iterator[None]:
    """Context manager that attaches extra information to exceptions."""
    try:
        yield

    except BadParameter as error:
        if error.ctx is None:
            error.ctx = ctx
        if param is not None and error.param is None:
            error.param = param
        raise

    except (Help, UsageError) as error:
        if error.ctx is None:
            error.ctx = ctx
        raise

    except parameter_errors as error:
        raise BadParameter(str(error), ctx, param) from error

    except usage_errors as error:
        raise UsageError(str(error), ctx) from error


class Abort(RuntimeError):  # noqa: N818
    pass


class Exit(RuntimeError):  # noqa: N818
    """An exception that indicates that the application should exit with some
    status code.

    :param code: the status code to exit with.
    """

    __slots__ = ("exit_code",)

    def __init__(self, code: int = 0) -> None:
        self.exit_code = code


class TabbError(Exception):
    """An exception that TaBB can handle and show to the user."""

    #: The exit code for this exception.
    exit_code = 1

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message

    def format_message(self) -> str:
        return self.message

    def __str__(self) -> str:
        return self.message

    def show(self, file: TextIO = sys.stderr) -> None:
        print(f"Error: {self.format_message()}", file=file)


class Help(TabbError):  # noqa: N818
    """An internal exception that signals TaBB to show the help page."""

    def __init__(
        self, message: str | None = None, ctx: Context[Any] | None = None
    ) -> None:
        super().__init__(message or "Help requested.")
        self.ctx = ctx

    def show(self, file: TextIO = sys.stderr) -> None:
        if self.ctx is None:
            raise RuntimeError("Cannot show help without ctx.")

        print(self.ctx.get_help(), file=file)


class UsageError(TabbError):
    """An internal exception that signals a usage error. This typically
    aborts any further handling."""

    exit_code = 2

    def __init__(self, message: str, ctx: Context[Any] | None = None) -> None:
        super().__init__(message)
        self.ctx = ctx

    def show(self, file: TextIO = sys.stderr) -> None:
        if self.ctx is not None:
            print(self.ctx.get_usage(), file=file)

            help_option = self.ctx.command.get_help_option(self.ctx)
            if help_option is not None:
                print(
                    f"Try '{self.ctx.command_path} {help_option.flags[0]}' for help.",
                    file=file,
                )

        super().show(file=file)


class BadParameter(UsageError):  # noqa: N818
    """An exception that formats out a standardized error message for a
    bad parameter.  This is useful when thrown from a callback or type as
    TaBB will attach contextual information to it (for instance, which
    parameter it is).
    """

    def __init__(
        self,
        message: str,
        ctx: Context[Any] | None = None,
        param: ParserParameter[Any] | None = None,
        param_hint: str | None = None,
    ) -> None:
        super().__init__(message, ctx)
        self.param = param
        self.param_hint = param_hint

    def format_message(self) -> str:
        if self.param_hint is not None:
            param_hint = self.param_hint
        elif self.param is not None:
            param_hint = self.param.get_error_hint(self.ctx)
        else:
            return f"Invalid value: {self.message}"

        return f"Invalid value for {param_hint}: {self.message}"


class MissingParameter(BadParameter):
    """Raised if a required an option or argument is missing."""

    def __init__(
        self,
        message: str | None = None,
        ctx: Context[Any] | None = None,
        param: ParserParameter[Any] | None = None,
        param_hint: str | None = None,
        param_type: str = "parameter",
    ) -> None:
        super().__init__(message or "", ctx, param, param_hint)
        self.param_type = param_type

    def _get_param_hint(self) -> str | None:
        if self.param_hint is not None:
            return self.param_hint

        if self.param is not None:
            return self.param.get_error_hint(self.ctx)

        return None

    def format_message(self) -> str:
        msg = f"Missing {self.param_type}"

        param_hint = self._get_param_hint()
        if param_hint:
            msg = f"{msg} {param_hint}"

        if self.message:
            return f"{msg}: {self.message}"

        return f"{msg}."

    def __str__(self) -> str:
        if self.message:
            return self.message

        if self.param:
            return f"Missing {self.param_type}: {self.param.name}"

        return f"Missing {self.param_type}."


class UnexpectedParameter(UsageError):  # noqa: N818
    def __init__(
        self,
        arg: str,
        message: str | None = None,
        possibilities: Sequence[str] | None = None,
        ctx: Context[Any] | None = None,
    ) -> None:
        if message is None:
            message = f"Unexpeted parameter: {arg}"

        super().__init__(message, ctx)
        self.arg = arg
        self.possibilities = possibilities

    def format_message(self) -> str:
        if not self.possibilities:
            return self.message

        if len(self.possibilities) == 1:
            suggest = f"Did you mean {self.possibilities[0]}?"
        else:
            options = ", ".join(sorted(self.possibilities))
            suggest = f"(Possible options: {options})"

        return f"{self.message} {suggest}"



from __future__ import annotations

import abc
import dataclasses
import operator
import os
import pathlib
import sys
import typing
from collections.abc import Callable, Hashable, Mapping, Sequence, Sized
from os import PathLike
from types import UnionType
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Generic,
    NoReturn,
    TypeAlias,
    TypeGuard,
    TypeVar,
    Union,
    cast,
    overload,
)

from tabb.exceptions import BadParameter, Help
from tabb.missing import _MISSING_TYPE, MISSING
from tabb.nargs import IntNArgs, NArgs, NArgsLiteral, VariadicNArgs
from tabb.utils import safecall, split_csv

if TYPE_CHECKING:
    from tabb.context import Context
    from tabb.parameter import ParserParameter

KeyType = TypeVar("KeyType", bound=Hashable)
SizedValueType = TypeVar("SizedValueType", bound=Sized)
ValueType = TypeVar("ValueType")

Number = TypeVar("Number", int, float)
NumberLike = TypeVar("NumberLike", bound=int | float)


@dataclasses.dataclass(slots=True, frozen=True)
class PositionalArg:
    value: str


@dataclasses.dataclass(slots=True, frozen=True)
class OptionArg:
    flag: str
    value: str | None


ParameterArg: TypeAlias = PositionalArg | OptionArg


class ParameterType(abc.ABC, Generic[ValueType]):
    @abc.abstractproperty
    def name(self) -> str:
        pass

    @property
    def nargs(self) -> NArgs:
        return IntNArgs(1)

    def has_default(self) -> bool:
        return_type = typing.get_type_hints(self.process_args).get("return", None)
        if return_type is _MISSING_TYPE:
            return False

        if typing.get_origin(return_type) not in (Union, UnionType):
            return True

        return _MISSING_TYPE not in typing.get_args(return_type)

    def get_metavar(self, ctx: Context[Any] | None) -> str | None:
        """Returns the metavar default for this param if it provides one."""
        return None

    def parse_flags(self, flags: Sequence[str]) -> tuple[list[str], list[str]]:
        return list(flags), []

    def format_value(self, value: ValueType) -> str:
        return repr(value)

    def fail(
        self,
        message: str,
        ctx: Context[Any] | None = None,
        param: ParserParameter[Any] | None = None,
    ) -> NoReturn:
        """Helper method to fail with an invalid value message."""
        raise BadParameter(message, ctx=ctx, param=param)

    def process_config(self, value: object) -> object:
        return value

    @abc.abstractmethod
    def matches(self, arg: ParameterArg) -> bool:
        """Reduces a new argument value into the current value."""
        ...

    @abc.abstractmethod
    def process_args(self, args: list[ParameterArg]) -> ValueType | _MISSING_TYPE:
        """Reduces a new argument value into the current value."""
        ...

    @abc.abstractmethod
    def parse_envvar(self, value: str) -> ValueType:
        """Parses a value from an environment variable."""
        ...

    @abc.abstractmethod
    def validate(self, value: Any) -> TypeGuard[ValueType]:
        """Validate a value after it was parsed."""
        ...


class TypeAdapter(ParameterType[ValueType]):
    def __init__(self, type: ParameterType[ValueType]) -> None:
        self.type = type
        super().__init__()

    @property
    def name(self) -> str:
        return f"<{type(self).__name__} type={self.type.name}>"

    @property
    def nargs(self) -> NArgs:
        return self.type.nargs

    def has_default(self) -> bool:
        return self.type.has_default()

    def get_metavar(self, ctx: Context[Any] | None) -> str | None:
        return self.type.get_metavar(ctx)

    def parse_flags(self, flags: Sequence[str]) -> tuple[list[str], list[str]]:
        return self.type.parse_flags(flags)

    def format_value(self, value: ValueType) -> str:
        return self.type.format_value(value)

    def matches(self, arg: ParameterArg) -> bool:
        return self.type.matches(arg)

    def process_args(self, args: list[ParameterArg]) -> ValueType | _MISSING_TYPE:
        return self.type.process_args(args)

    def process_config(self, value: object) -> object:
        return self.type.process_config(value)

    def parse_envvar(self, value: str) -> ValueType:
        return self.type.parse_envvar(value)

    def validate(self, value: Any) -> TypeGuard[ValueType]:
        return self.type.validate(value)


class AnyType(ParameterType[Any]):
    @property
    def name(self) -> str:
        return "Any"

    @property
    def nargs(self) -> NArgs:
        return IntNArgs(0)

    def has_default(self) -> bool:
        return False

    def matches(self, arg: ParameterArg) -> bool:
        return False

    def process_args(self, args: list[ParameterArg]) -> Any | _MISSING_TYPE:
        return MISSING

    def parse_envvar(self, value: str) -> Any:
        return value

    def validate(self, value: Any) -> TypeGuard[Any]:
        """Validate a value after it was parsed."""
        return True


class Scalar(ParameterType[ValueType]):
    def __init__(
        self, type: Callable[[str], ValueType], allow_overwrite: bool = True
    ) -> None:
        self.type = type
        self.allow_overwrite = allow_overwrite
        super().__init__()

    @property
    def name(self) -> str:
        # Return name if type is builtin
        if self.type.__module__ == "builtins":
            return self.type.__name__.lower()
        return repr(self.type)

    def matches(self, arg: ParameterArg) -> bool:
        if arg.value is None:
            return False

        try:
            self.type(arg.value)
        except (BadParameter, ValueError):
            return False

        return True

    def parse(self, value: str) -> ValueType:
        try:
            return self.type(value)
        except ValueError:
            self.fail(f"{value!r} is not a valid {self.name}.")

    def process_arg(self, arg: ParameterArg) -> ValueType:
        if arg.value is None:
            raise TypeError("Expected string value")
        return self.parse(arg.value)

    def process_args(self, args: list[ParameterArg]) -> ValueType | _MISSING_TYPE:
        if not args:
            return MISSING

        if len(args) != 1 and not self.allow_overwrite:
            self.fail("Parameter already set.")

        return self.process_arg(args[-1])

    def parse_envvar(self, value: str) -> ValueType:
        return self.parse(value)

    def validate(self, value: Any) -> TypeGuard[ValueType]:
        if isinstance(self.type, type) and not isinstance(value, self.type):
            self.fail(f"Expected {self.name} value")
        return True


class String(Scalar[str]):
    def __init__(self, allow_overwrite: bool = True) -> None:
        super().__init__(str, allow_overwrite)


class Int(Scalar[int]):
    def __init__(self, allow_overwrite: bool = True) -> None:
        super().__init__(int, allow_overwrite)


class Float(Scalar[float]):
    def __init__(self, allow_overwrite: bool = True) -> None:
        super().__init__(float, allow_overwrite)


class Bool(Scalar[bool]):
    true_values = ("1", "true", "t", "yes", "y", "on")
    false_values = ("0", "false", "f", "no", "n", "off")

    def __init__(self, allow_overwrite: bool = False) -> None:
        super().__init__(bool, allow_overwrite=allow_overwrite)

    @property
    def name(self) -> str:
        return "bool"

    def format_value(self, value: bool) -> str:
        return "true" if value else "false"

    def matches(self, arg: ParameterArg) -> bool:
        if arg.value is None:
            return False
        value = arg.value.lower()
        return value in self.true_values or value in self.false_values

    def parse(self, value: str) -> bool:
        if value.lower() in self.true_values:
            return True

        if value.lower() in self.false_values:
            return False

        self.fail(f"{value!r} is not a valid bool (try true/false).")


class Flag(Bool):
    def __init__(self, allow_overwrite: bool = False) -> None:
        self.true_flags: list[str] = []
        self.false_flags: list[str] = []
        super().__init__(allow_overwrite=allow_overwrite)

    @property
    def name(self) -> str:
        return "<BooleanFlag>"

    @property
    def nargs(self) -> NArgs:
        return IntNArgs(0)

    def format_value(self, value: bool) -> str:
        return self.true_flags[0] if value else self.false_flags[0]

    def parse_flags(self, flags: Sequence[str]) -> tuple[list[str], list[str]]:
        for flag in flags:
            true_flag, _, false_flag = flag.partition("/")

            if true_flag:
                self.true_flags.append(true_flag)

            if false_flag:
                self.false_flags.append(false_flag)

            elif true_flag.startswith("--"):
                self.false_flags.append(f"--no-{true_flag[2:]}")

        if set(self.true_flags).intersection(self.false_flags):
            raise ValueError("True and false flags overlap")

        return self.true_flags, self.false_flags

    def matches(self, arg: ParameterArg) -> bool:
        if not isinstance(arg, OptionArg):
            return False

        if arg.value is not None:
            return False

        return arg.flag in self.true_flags or arg.flag in self.false_flags

    def process_arg(self, arg: ParameterArg) -> bool:
        if arg.value is not None:
            raise TypeError("Expected None value")

        if arg.flag in self.true_flags:
            return True

        if arg.flag in self.false_flags:
            return False

        msg = f"Unexpected flag: {arg.flag!r}"
        raise RuntimeError(msg)


class Counter(ParameterType[Number]):
    def __init__(self, value_type: type[Number] = int) -> None:
        self.value_type: type[Number] = value_type
        super().__init__()

    @property
    def name(self) -> str:
        return f"<Counter type={self.value_type.__name__}>"

    @property
    def nargs(self) -> NArgs:
        return IntNArgs(0)

    def matches(self, arg: ParameterArg) -> bool:
        return arg.value is None

    def process_args(self, args: list[ParameterArg]) -> Number:
        return self.value_type(len(args))

    def parse_envvar(self, value: str) -> Number:
        try:
            return self.value_type(value)
        except ValueError:
            self.fail(f"{value!r} is not a valid {self.value_type.__name__}.")

    def validate(self, value: Any) -> TypeGuard[Number]:
        if not isinstance(value, self.value_type):
            self.fail(f"Expected {self.value_type.__name__} value")
        return True


class Optional(TypeAdapter[ValueType | None]):
    @property
    def name(self) -> str:
        return f"Optional[{self.type.name}]"

    def process_config(self, value: object) -> object | None:
        if value is None:
            return None
        return super().process_config(value)

    def validate(self, value: Any) -> TypeGuard[ValueType | None]:
        if value is None:
            return True
        return super().validate(value)


class SequenceType(ParameterType[Sequence[ValueType]]):
    def __init__(
        self,
        item_type: ParameterType[ValueType],
        *,
        nargs: NArgs | NArgsLiteral | None = None,
    ) -> None:
        if nargs is not None and not isinstance(nargs, NArgs):
            nargs = NArgs.from_literal(nargs)

        self.item_type = item_type
        self._nargs = nargs
        super().__init__()

    @property
    def name(self) -> str:
        return f"Sequence[{self.item_type.name}]"

    @property
    def nargs(self) -> NArgs:
        if self._nargs is None:
            return self.item_type.nargs
        return self._nargs

    def get_metavar(self, ctx: Context[Any] | None) -> str | None:
        return self.item_type.get_metavar(ctx)

    def parse_flags(self, flags: Sequence[str]) -> tuple[list[str], list[str]]:
        return self.item_type.parse_flags(flags)

    def format_value(self, value: Sequence[ValueType]) -> str:
        return ", ".join(self.item_type.format_value(item) for item in value)

    def matches(self, arg: ParameterArg) -> bool:
        return self.item_type.matches(arg)

    def process_args(self, args: list[ParameterArg]) -> Sequence[ValueType]:
        result: list[ValueType] = []

        for arg in args:
            value = self.item_type.process_args([arg])
            if isinstance(value, _MISSING_TYPE):
                continue
            result.append(value)

        return result

    def process_config(self, value: object) -> object:
        if not isinstance(value, Sequence):
            return value
        return [self.item_type.process_config(item) for item in value]

    def parse_envvar(self, value: str) -> Sequence[ValueType]:
        return [self.item_type.parse_envvar(s) for s in split_csv(value)]

    def validate(self, value: Any) -> TypeGuard[Sequence[ValueType]]:
        if self.validate_type(value):
            return self.validate_items(value)
        return False

    def validate_type(self, value: Any) -> TypeGuard[Sequence[Any]]:
        if not isinstance(value, Sequence):
            self.fail("Expected Sequence")
        return True

    def validate_items(self, value: Sequence[Any]) -> TypeGuard[Sequence[ValueType]]:
        for item in value:
            self.item_type.validate(item)
        return True


class List(SequenceType[ValueType], ParameterType[list[ValueType]]):
    def _cast(self, value: Sequence[ValueType]) -> list[ValueType]:
        if isinstance(value, list):
            return value
        return list(value)

    @property
    def name(self) -> str:
        return f"list[{self.item_type.name}]"

    def format_value(self, value: Sequence[ValueType]) -> str:
        return f"[{super().format_value(value)}]"

    def process_args(self, args: list[ParameterArg]) -> list[ValueType]:
        return self._cast(super().process_args(args))

    def process_config(self, value: object) -> object:
        value = super().process_config(value)
        if isinstance(value, Sequence):
            return self._cast(value)
        return value

    def parse_envvar(self, value: str) -> list[ValueType]:
        return self._cast(super().parse_envvar(value))

    def validate_type(self, value: Any) -> TypeGuard[list[Any]]:
        if not isinstance(value, list):
            self.fail("Expected list")
        return True

    def validate(self, value: Any) -> TypeGuard[list[ValueType]]:
        return super().validate(value)


class Tuple(SequenceType[ValueType], ParameterType[list[ValueType]]):
    def _cast(self, value: Sequence[ValueType]) -> tuple[ValueType, ...]:
        return tuple(value)

    @property
    def name(self) -> str:
        return f"list[{self.item_type.name}]"

    def format_value(self, value: Sequence[ValueType]) -> str:
        return f"[{super().format_value(value)}]"

    def process_args(self, args: list[ParameterArg]) -> tuple[ValueType, ...]:
        return self._cast(super().process_args(args))

    def process_config(self, value: object) -> object:
        value = super().process_config(value)
        if isinstance(value, Sequence):
            return self._cast(value)
        return value

    def parse_envvar(self, value: str) -> tuple[ValueType, ...]:
        return self._cast(super().parse_envvar(value))

    def validate_type(self, value: Any) -> TypeGuard[tuple[Any, ...]]:
        if not isinstance(value, tuple):
            self.fail("Expected tuple")
        return True

    def validate(self, value: Any) -> TypeGuard[tuple[ValueType, ...]]:
        return super().validate(value)


class Dict(ParameterType[dict[KeyType, ValueType]]):
    @overload
    def __init__(
        self: Dict[str, ValueType],
        value_type: ParameterType[ValueType],
        *,
        nargs: NArgs | NArgsLiteral | None = None,
    ) -> None:
        ...

    @overload
    def __init__(
        self: Dict[KeyType, ValueType],
        value_type: ParameterType[ValueType],
        key_type: ParameterType[KeyType],
        *,
        nargs: NArgs | NArgsLiteral | None = None,
    ) -> None:
        ...

    def __init__(
        self,
        value_type: ParameterType[ValueType],
        key_type: ParameterType[KeyType] | None = None,
        *,
        nargs: NArgs | NArgsLiteral | None = None,
    ) -> None:
        if key_type is None:
            key_type = cast(ParameterType[KeyType], Scalar(str))

        if key_type.nargs != IntNArgs(1):
            raise ValueError("DictType does not support nargs other than 1")

        if value_type.nargs != IntNArgs(1):
            raise ValueError("DictType does not support nargs other than 1")

        if nargs is not None and not isinstance(nargs, NArgs):
            nargs = NArgs.from_literal(nargs)

        self.value_type = value_type
        self.key_type = key_type
        self._nargs = nargs
        super().__init__()

    @property
    def name(self) -> str:
        return f"dict[{self.key_type.name}, {self.value_type.name}]"

    @property
    def nargs(self) -> NArgs:
        if self._nargs is None:
            return IntNArgs(1)
        return self._nargs

    def get_metavar(self, ctx: Context[Any] | None) -> str | None:
        key_metavar = self.key_type.get_metavar(ctx) or "KEY"
        value_metavar = self.value_type.get_metavar(ctx) or "VALUE"
        return f"{key_metavar}={value_metavar}"

    def parse_flags(self, flags: Sequence[str]) -> tuple[list[str], list[str]]:
        return self.value_type.parse_flags(flags)

    def format_value(self, value: dict[KeyType, ValueType]) -> str:
        return ", ".join(
            f"{self.key_type.format_value(k)}={self.value_type.format_value(v)}"
            for k, v in value.items()
        )

    def parse_arg(self, arg: ParameterArg) -> tuple[ParameterArg, ParameterArg] | None:
        if not isinstance(arg, OptionArg):
            return None

        if arg.value is None:
            return None

        key, has_sep, value = arg.value.partition("=")
        if not has_sep:
            return None

        if isinstance(arg, PositionalArg):
            return PositionalArg(value=key), PositionalArg(value=value)

        return (
            OptionArg(flag=arg.flag, value=key),
            OptionArg(flag=arg.flag, value=value),
        )

    def matches(self, arg: ParameterArg) -> bool:
        if arg.value is None:
            return False

        item = self.parse_arg(arg)
        if item is None:
            return False

        return self.key_type.matches(item[0]) and self.value_type.matches(item[0])

    def process_args(self, args: list[ParameterArg]) -> dict[KeyType, ValueType]:
        def _process(arg: ParameterArg) -> tuple[KeyType, ValueType] | None:
            item = self.parse_arg(arg)
            if item is None:
                return None

            key_result = self.key_type.process_args([item[0]])
            if isinstance(key_result, _MISSING_TYPE):
                return None

            value_result = self.value_type.process_args([item[1]])
            if isinstance(value_result, _MISSING_TYPE):
                return None

            return (key_result, value_result)

        return dict(item for arg in args if (item := _process(arg)))

    def process_config(self, value: object) -> object:
        if not isinstance(value, Mapping):
            return value

        def _process(key: object, value: object) -> tuple[object, object]:
            return (
                self.key_type.process_config(key),
                self.value_type.process_config(value),
            )

        return dict(_process(k, v) for k, v in value.items())

    def parse_envvar(self, value: str) -> dict[KeyType, ValueType]:
        def _parse(item: str) -> tuple[KeyType, ValueType]:
            k, has_sep, v = item.partition("=")
            if not has_sep:
                self.fail(f"Expected key=value, got {item!r}")

            return (
                self.key_type.parse_envvar(k),
                self.value_type.parse_envvar(v),
            )

        return dict(_parse(item) for item in split_csv(value))

    def validate(self, value: Any) -> TypeGuard[dict[KeyType, ValueType]]:
        if not isinstance(value, dict):
            self.fail("Expected dict")

        for key, item in value.items():
            self.key_type.validate(key)
            self.value_type.validate(item)

        return True


class Path(Scalar[pathlib.Path]):
    @property
    def name(self) -> str:
        return "path"

    def __init__(
        self,
        *,
        allow_dash: bool = False,
        allow_overwrite: bool = True,
        dir_okay: bool = True,
        executable: bool = False,
        exists: bool = False,
        file_okay: bool = True,
        readable: bool = False,
        resolve_path: bool = False,
        writable: bool = False,
    ) -> None:
        super().__init__(pathlib.Path, allow_overwrite)

        self.allow_dash = allow_dash
        self.dir_okay = dir_okay
        self.executable = executable
        self.exists = exists
        self.file_okay = file_okay
        self.readable = readable
        self.resolve_path = resolve_path
        self.writable = writable

    def parse(self, value: str) -> pathlib.Path:
        result = super().parse(value)

        if self.resolve_path and value != "-":
            result = result.resolve()

        return result

    def process_config(self, value: object) -> object:
        if not isinstance(value, str):
            return value
        return self.parse(value)

    def _validate_dash(self) -> bool:
        if not self.allow_dash:
            self.fail("'-' is not allowed.")

        if self.executable:
            self.fail("Path must be executable.")

        return True

    def validate(self, value: Any) -> TypeGuard[pathlib.Path]:
        if not isinstance(value, pathlib.Path):
            self.fail("Expected path value.")

        if value == pathlib.Path("-"):
            return self._validate_dash()

        if self.exists and not value.exists():
            self.fail("Path must exist.")

        if not self.file_okay and value.is_file():
            self.fail("Path must not be a file.")

        if not self.dir_okay and value.is_dir():
            self.fail("Path must not be a directory.")

        if self.readable and not os.access(value, os.R_OK):
            self.fail("Path must be readable.")

        if self.writable and not os.access(value, os.W_OK):
            self.fail("Path must be writable.")

        if self.executable and not os.access(value, os.X_OK):
            self.fail("Path must be executable.")

        return True


class File(ParameterType[IO[Any]]):
    def __init__(
        self,
        mode: str = "r",
        encoding: str = "utf-8",
        errors: str = "strict",
        default: int | str | bytes | PathLike[str] | PathLike[bytes] | None = None,
        allow_overwrite: bool = True,
        close: bool | None = None,
    ) -> None:
        super().__init__()
        self.mode = mode
        self.encoding = encoding
        self.errors = errors
        self.default = default
        self.allow_overwrite = allow_overwrite
        self.close = close

    @property
    def name(self) -> str:
        return f"File[{self.mode!r}]"

    def format_value(self, value: IO[Any]) -> str:
        if hasattr(value, "name"):
            if value.name in ("<stdin>", "<stdout>"):
                return "-"

            return value.name

        return repr(value)

    def has_default(self) -> bool:
        return self.default is not None

    def matches(self, arg: ParameterArg) -> bool:
        return arg.value is not None

    def open(self, value: str) -> IO[Any]:
        if os.fsdecode(value) == "-":
            if any(m in self.mode for m in ["w", "a", "x"]):
                if "b" in self.mode:
                    return sys.stdout.buffer
                return sys.stdout

            if "b" in self.mode:
                return sys.stdin.buffer
            return sys.stdin

        result = open(value, self.mode, encoding=self.encoding, errors=self.errors)

        from tabb.context import get_current_context

        ctx = get_current_context(silent=True)

        if ctx is not None:
            ctx.call_on_close(safecall(result.close))

        return result

    def process_args(self, args: list[ParameterArg]) -> IO[Any] | _MISSING_TYPE:
        if not args:
            if not self.default:
                return MISSING
            return self.open(self.default)

        if len(args) != 1 and not self.allow_overwrite:
            self.fail("Parameter already set.")

        return self.open(args[-1])

    def process_config(self, value: object) -> IO[Any]:
        if not isinstance(value, str):
            self.fail("Config value must be a string.")
        return self.open(value)

    def parse_envvar(self, value: str) -> IO[Any]:
        return self.open(value)

    def validate(self, value: Any) -> TypeGuard[IO[Any]]:
        if not (hasattr(value, "read") or hasattr(value, "write")):
            self.fail("Expected file-like object.")
        return True


class Validator(TypeAdapter[ValueType]):
    def __init__(
        self,
        type: ParameterType[ValueType],
        validator: Callable[[ValueType], bool],
    ) -> None:
        self.validator = validator
        super().__init__(type)

    @property
    def name(self) -> str:
        return f"<Validator type={self.type.name} validator={self.validator!r}>"

    def validate(self, value: Any) -> TypeGuard[ValueType]:
        return self.type.validate(value) and self.validator(value)


class ChoicesValidator(TypeAdapter[ValueType]):
    def __init__(
        self, type: ParameterType[ValueType], choices: Sequence[ValueType]
    ) -> None:
        self.choices = list(choices)
        super().__init__(type)

    @property
    def name(self) -> str:
        choices = "|".join(str(choice) for choice in self.choices)
        return f"<ChoicesValidator type={self.type.name} choices={choices!r}>"

    def get_metavar(self, ctx: Context[Any] | None) -> str | None:
        return "|".join(self.type.format_value(choice) for choice in self.choices)

    def validate(self, value: Any) -> TypeGuard[ValueType]:
        self.type.validate(value)

        if value not in self.choices:
            choices = ", ".join(str(choice) for choice in self.choices)
            self.fail(f"{value!r} is not one of {choices}.")

        return True


class RangeValidator(TypeAdapter[NumberLike]):
    def __init__(
        self,
        type: ParameterType[NumberLike],
        cast: Callable[[int | float | str], NumberLike],
        min: int | float | None = None,
        max: int | float | None = None,
        min_open: bool = False,
        max_open: bool = False,
        clamp: bool = False,
    ) -> None:
        self.cast = cast
        self.min = min
        self.max = max
        self.min_open = min_open
        self.max_open = max_open
        self.clamp = clamp
        super().__init__(type)

    @property
    def name(self) -> str:
        return (
            f"<NumberRangeValidator type={self.type.name} "
            "min={self.min} max={self.max} "
            "min_open={self.min_open} max_open={self.max_open} "
            "clamp={self.clamp}>"
        )

    def lt_min(self, value: NumberLike) -> bool:
        if self.min is None:
            return False
        lt = operator.le if self.min_open else operator.lt
        return bool(lt(value, self.min))

    def gt_max(self, value: NumberLike) -> bool:
        if self.max is None:
            return False
        gt = operator.ge if self.max_open else operator.gt
        return bool(gt(value, self.max))

    def maybe_clamp(self, value: NumberLike) -> NumberLike:
        if not self.clamp:
            return value

        if self.min is not None:
            value = self.cast(max(self.min, value))

        if self.max is not None:
            value = self.cast(min(self.max, value))

        return value

    def process_args(self, args: list[ParameterArg]) -> NumberLike | _MISSING_TYPE:
        result = self.type.process_args(args)
        if isinstance(result, _MISSING_TYPE):
            return MISSING
        return self.maybe_clamp(result)

    def parse_envvar(self, value: str) -> NumberLike:
        result = self.type.parse_envvar(value)
        return self.maybe_clamp(result)

    def validate(self, value: Any) -> TypeGuard[NumberLike]:
        if self.type.validate(value):
            if self.lt_min(value):
                self.fail(f"{value!r} is smaller than the minimum value of {self.min}.")

            if self.gt_max(value):
                self.fail(f"{value!r} is greater than the maximum value of {self.max}.")
            return True
        return False


class LengthValidator(TypeAdapter[SizedValueType]):
    def __init__(
        self,
        type: ParameterType[SizedValueType],
        min_length: int | None = None,
        max_length: int | None = None,
    ) -> None:
        self.min_length = min_length
        self.max_length = max_length
        super().__init__(type)

    @property
    def name(self) -> str:
        return (
            f"<LengthValidator type={self.type.name} "
            "min_length={self.min_length} max_length={self.max_length}>"
        )

    @property
    def nargs(self) -> NArgs:
        child_nargs: NArgs = self.type.nargs

        if not isinstance(child_nargs, VariadicNArgs):
            return child_nargs

        min_values = child_nargs.min_values
        if self.min_length is not None:
            min_values = max(min_values, self.min_length)

        max_values = child_nargs.max_values
        if self.max_length is not None:
            if max_values is None:
                max_values = self.max_length
            else:
                max_values = min(max_values, self.max_length)

        return VariadicNArgs(
            value=child_nargs.value,
            min_values=min_values,
            max_values=max_values,
            greedy=child_nargs.greedy,
        )

    def validate(self, value: Any) -> TypeGuard[SizedValueType]:
        if self.type.validate(value):
            if self.min_length is not None and len(value) < self.min_length:
                self.fail(
                    f"{value!r} is shorter than the minimum length of "
                    f"{self.min_length}."
                )

            if self.max_length is not None and len(value) > self.max_length:
                self.fail(
                    f"{value!r} is longer than the maximum length of "
                    f"{self.max_length}."
                )

            return True
        return False


class GreedyAdaptor(TypeAdapter[ValueType,]):
    def __init__(self, type: ParameterType[ValueType,], greedy: bool = True) -> None:
        self.greedy = greedy
        super().__init__(type)

    @property
    def name(self) -> str:
        return f"<GreedyAdaptor type={self.type.name} greedy={self.greedy}>"

    @property
    def nargs(self) -> NArgs:
        child_nargs: NArgs = self.type.nargs

        if not isinstance(child_nargs, VariadicNArgs):
            return child_nargs

        return VariadicNArgs(
            value=child_nargs.value,
            min_values=child_nargs.min_values,
            max_values=child_nargs.max_values,
            greedy=self.greedy,
        )


class HelpFlag(ParameterType[NoReturn]):
    @property
    def name(self) -> str:
        return "<Help>"

    @property
    def nargs(self) -> NArgs:
        return IntNArgs(0)

    def matches(self, arg: ParameterArg) -> bool:
        return arg.value is None

    def process_args(self, args: list[ParameterArg]) -> NoReturn | _MISSING_TYPE:
        if not args:
            return MISSING
        raise Help()

    def parse_envvar(self, value: str) -> NoReturn:
        raise Help()

    def validate(self, value: Any) -> TypeGuard[NoReturn]:
        raise Help()


# TODO: add types:
# - url
# - uuid
# - enum



from __future__ import annotations

import inspect
import typing
from collections.abc import Callable
from typing import Any, TypeVar

from tabb.context import Context
from tabb.parameter.base import (
    Parameter,
    ParameterKind,
    ParameterSource,
    ParameterValue,
)
from tabb.parameter.utils import get_base_type

ValueType = TypeVar("ValueType")


class DependsParameter(Parameter[ValueType]):
    def __init__(
        self,
        name: str,
        kind: ParameterKind,
        dependency: Callable[..., ValueType],
        *,
        use_cache: bool = True,
    ) -> None:
        super().__init__(name, kind)
        self.dependency = dependency
        self.use_cache = use_cache

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.dependency.__name__}>"

    def get_sub_params(self, ctx: Context[Any]) -> list[Parameter[Any]]:
        from tabb.callback import Callback

        callback = self.get_callback(ctx)

        if isinstance(callback, Callback):
            return callback.get_params(ctx)

        return []

    def resolve(self, ctx: Context[Any]) -> ParameterValue[ValueType]:
        return ParameterValue(self.get_value(ctx), ParameterSource.DEPENDS, self)

    def make_callback(self) -> Callable[[Context[Any]], ValueType]:
        signature = inspect.signature(self.dependency)
        parameters = list(signature.parameters.values())

        if len(parameters) == 1:
            type_hints = typing.get_type_hints(self.dependency, include_extras=False)
            type_hint = type_hints[parameters[0].name]
            if get_base_type(type_hint) is Context:
                return self.dependency

        from tabb.callback import Callback

        return Callback(self.dependency)

    def get_callback(self, ctx: Context[Any]) -> Callable[[Context[Any]], ValueType]:
        if self.dependency not in ctx.dependencies.callbacks:
            ctx.dependencies.callbacks[self.dependency] = self.make_callback()
        return ctx.dependencies.callbacks[self.dependency]

    def get_value(self, ctx: Context[Any]) -> ValueType:
        if not self.use_cache:
            dependency = self.get_callback(ctx)
            return dependency(ctx)

        if self.dependency not in ctx.dependencies.values:
            dependency = self.get_callback(ctx)
            ctx.dependencies.values[self.dependency] = dependency(ctx)

        value: ValueType = ctx.dependencies.values[self.dependency]
        return value



from __future__ import annotations

import typing
from types import NoneType, UnionType
from typing import Annotated, Literal, TypeVar, Union

T = TypeVar("T")


def get_annotations(type_hint: object) -> tuple[object, list[object]]:
    if typing.get_origin(type_hint) is Annotated:
        type_args = typing.get_args(type_hint)
        return type_args[0], list(type_args[1:])
    return type_hint, []


def get_annotated_child_type(type_hint: object) -> tuple[bool, object]:
    if typing.get_origin(type_hint) is not Annotated:
        return False, type_hint

    type_args = typing.get_args(type_hint)
    return True, type_args[0]


def get_optional_child_type(type_hint: object) -> tuple[bool, object]:
    if typing.get_origin(type_hint) not in (Union, UnionType):
        return False, type_hint

    type_args = typing.get_args(type_hint)
    if len(type_args) != 2 or NoneType not in type_args:
        msg = "Union type arguments not supported."
        raise TypeError(msg)

    return True, next(t for t in type_args if t is not NoneType)


def get_list_child_type(type_hint: object) -> tuple[bool, object]:
    if typing.get_origin(type_hint) is not list:
        return False, type_hint

    type_args = typing.get_args(type_hint)
    if len(type_args) != 1:
        msg = "List type arguments should contain a single type."
        raise TypeError(msg)

    return True, type_args[0]


def get_tuple_child_type(type_hint: object) -> tuple[bool, object]:
    if typing.get_origin(type_hint) is not tuple:
        return False, type_hint

    type_args = typing.get_args(type_hint)
    if len(type_args) != 2 or type_args[1] != ...:
        msg = "Tuple type arguments should contain a single type and ellipsis."
        raise TypeError(msg)

    return True, type_args[0]


def get_literal_child_type(type_hint: object) -> tuple[bool, object]:
    if typing.get_origin(type_hint) is not Literal:
        return False, type_hint

    type_args = typing.get_args(type_hint)
    # Make sure just a single type has been used
    if len({type(arg) for arg in type_args}) != 1:
        msg = "Literal type arguments should contain iterms of a single type."
        raise TypeError(msg)

    return True, type(type_args[0])


def _strip_typing_extra(type_hint: object) -> object:
    if isinstance(origin := typing.get_origin(type_hint), type):
        return origin

    for resolvers in (
        get_annotated_child_type,
        get_optional_child_type,
        get_literal_child_type,
    ):
        found, child_type = resolvers(type_hint)
        if found:
            return child_type

    msg = f"Unexpted type {type_hint!r}."
    raise TypeError(msg)


def get_base_type(type_hint: object) -> type:
    while not isinstance(type_hint, type):
        type_hint = _strip_typing_extra(type_hint)
    return type_hint



from __future__ import annotations

import os
import sys
from collections.abc import Callable
from functools import update_wrapper
from shlex import shlex
from types import ModuleType
from typing import Any, ParamSpec, TextIO, TypeVar, cast

P = ParamSpec("P")
R = TypeVar("R")
D = TypeVar("D")


def to_kebab(name: str) -> str:
    return name.translate(str.maketrans("_ ", "--"))


def to_snake(name: str) -> str:
    return name.translate(str.maketrans("- ", "__"))


def split_csv(value: str) -> list[str]:
    splitter = shlex(value, posix=True)
    splitter.whitespace = ","
    splitter.whitespace_split = True
    return [s.strip() for s in splitter]


def detect_program_name(
    path: str | None = None, main_module: ModuleType | None = None
) -> str:
    if main_module is None:
        main_module = sys.modules["__main__"]

    if not path:
        path = sys.argv[0]

    # The value of __package__ indicates how Python was called. It may
    # not exist if a setuptools script is installed as an egg. It may be
    # set incorrectly for entry points created with pip on Windows.
    if getattr(main_module, "__package__", None) is None or (
        os.name == "nt"
        and main_module.__package__ == ""
        and not os.path.exists(path)
        and os.path.exists(f"{path}.exe")
    ):
        # Executed a file, like "python app.py".
        return os.path.basename(path)

    # Executed a module, like "python -m example".
    # Rewritten by Python from "-m script" to "/path/to/script.py".
    # Need to look at main module to determine how it was executed.
    name = os.path.splitext(os.path.basename(path))[0]

    py_module = main_module.__package__
    assert py_module is not None  # noqa: S101

    # A submodule like "example.cli".
    if name != "__main__":
        py_module = f"{py_module}.{name}"

    return f"python -m {py_module.lstrip('.')}"


class PacifyFlushWrapper:
    """This wrapper is used to catch and suppress BrokenPipeErrors resulting
    from ``.flush()`` being called on broken pipe during the shutdown/final-GC
    of the Python interpreter. Notably ``.flush()`` is always called on
    ``sys.stdout`` and ``sys.stderr``. So as to have minimal impact on any
    other cleanup code, and the case where the underlying file is not a broken
    pipe, all calls and attributes are proxied.
    """

    def __init__(self, wrapped: TextIO) -> None:
        self.wrapped = wrapped

    def flush(self) -> None:
        try:
            self.wrapped.flush()
        except OSError as e:
            import errno

            if e.errno != errno.EPIPE:
                raise

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.wrapped, attr)


def pacify_flush(file: TextIO) -> TextIO:
    """Return a wrapper around the given file that will suppress BrokenPipeErrors
    resulting from ``.flush()`` being called on a broken pipe.
    """
    return cast(TextIO, PacifyFlushWrapper(file))


def safecall(
    fn: Callable[P, R],
    default: D = None,
    exceptions: tuple[type[BaseException], ...] = (Exception,),
) -> Callable[P, R | D]:
    "Wraps a function so that it swallows exceptions."

    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R | D:
        try:
            return fn(*args, **kwargs)
        except exceptions:
            return default

    return update_wrapper(wrapper, fn)



from __future__ import annotations

import inspect
from collections.abc import Callable
from typing import Generic, ParamSpec, TypeVar

from tabb.base import BaseCommand
from tabb.callback import create_callback
from tabb.context import Context
from tabb.parameter import ParameterGroup
from tabb.utils import to_kebab

P = ParamSpec("P")
T = TypeVar("T")


class Command(BaseCommand[T], Generic[P, T]):
    def __init__(
        self,
        callback: str | Callable[P, T],
        *,
        add_help_option: bool = True,
        deprecated: bool = False,
        epilog: str | None = None,
        help: str | None = None,
        hidden: bool = False,
        import_dependencies: str | None = None,
        name: str | None = None,
        no_args_is_help: bool = False,
        options_metavar: str | None = "[OPTIONS]",
        short_help: str | None = None,
    ) -> None:
        super().__init__(
            add_help_option=add_help_option,
            deprecated=deprecated,
            epilog=epilog,
            help=help,
            hidden=hidden,
            name=name,
            no_args_is_help=no_args_is_help,
            options_metavar=options_metavar,
            short_help=short_help,
        )
        self._callback = create_callback(
            callback, import_dependencies=import_dependencies
        )

    def get_name(self) -> str:
        name = super().get_name()
        if name:
            return name
        return to_kebab(self._callback.name)

    def get_help_text(self) -> str:
        help_text = super().get_help_text()
        if help_text:
            return help_text

        if self._callback:
            return inspect.getdoc(self._callback.fn) or ""

        return ""

    def get_params(self, ctx: Context[T]) -> ParameterGroup:
        params = super().get_params(ctx)
        params.add_params(self._callback.get_params(ctx))
        return params

    def invoke(self, ctx: Context[T]) -> T:
        self.parse_args(ctx)
        self.invoke_callbacks(ctx)
        return self._callback(ctx)



from __future__ import annotations

from collections.abc import Iterator, Mapping
from typing import TypeVar, overload

U = TypeVar("U")


class Config(Mapping[str, object]):
    def __init__(self, *mappings: Mapping[str, object]) -> None:
        self._mappings: list[Mapping[str, object]] = list(mappings) if mappings else []

    def merge(self, data: Mapping[str, object]) -> None:
        self._mappings.append(data)

    def subconfig(self, key: str) -> Config:
        mappings: list[Mapping[str, object]] = []

        for mapping in self._mappings:
            try:
                value = mapping[key]
            except KeyError:
                continue

            if isinstance(value, Mapping):
                mappings.append(value)

        return self.__class__(*mappings)

    def __missing__(self, key: str) -> object:
        raise KeyError(key)

    def __getitem__(self, key: str) -> object:
        mappings = reversed(self._mappings)

        for mapping in mappings:
            try:
                value = mapping[key]
            except KeyError:
                continue

            if not isinstance(value, Mapping):
                return value

            submapping = self.__class__(value)

            for parent in mappings:
                try:
                    value = parent[key]
                except KeyError:
                    continue

                if not isinstance(value, Mapping):
                    return value

                submapping.merge(value)

            return submapping

        return self.__missing__(key)

    @overload
    def get(self, key: str) -> object | None:
        ...

    @overload
    def get(self, key: str, default: U) -> object | U:
        ...

    def get(self, key: str, default: U | None = None) -> object | U | None:
        return self[key] if key in self else default

    @overload
    def get_path(self, path: str) -> object | None:
        ...

    @overload
    def get_path(self, path: str, default: U) -> object | U:
        ...

    def get_path(self, path: str, default: U | None = None) -> object | U | None:
        keys = path.split(".")
        value: object = self

        for key in keys:
            if not isinstance(value, Mapping):
                return default

            try:
                value = value[key]
            except KeyError:
                return default

        return value

    def __len__(self) -> int:
        return len(set().union(*self._mappings))

    def __iter__(self) -> Iterator[str]:
        result = {}
        for mapping in reversed(self._mappings):
            result.update(dict.fromkeys(mapping))
        return iter(result)

    def __contains__(self, key: object) -> bool:
        if not isinstance(key, str):
            return False
        return any(key in m for m in self._mappings)

    def __bool__(self) -> bool:
        return any(self._mappings)

    def __repr__(self) -> str:
        mappings = ", ".join(map(repr, self._mappings))
        return f"{self.__class__.__name__}({mappings})"

    def copy(self) -> Config:
        return self.__class__(*self._mappings)

    __copy__ = copy



from __future__ import annotations

import abc
import dataclasses
import functools
import itertools
from collections.abc import Iterator, Sequence
from difflib import get_close_matches
from typing import Any, Generic, Literal, NoReturn, TypeVar

from tabb.context import Context
from tabb.exceptions import UnexpectedParameter
from tabb.nargs import IntNArgs, NArgs, VariadicNArgs
from tabb.parameter import (
    ArgumentParameter,
    DependsParameter,
    OptionParameter,
    ParameterValue,
    ParserParameter,
)
from tabb.parameter.types import OptionArg, ParameterArg, PositionalArg
from tabb.utils import to_snake

T = TypeVar("T")

CapturedArg = tuple[ParserParameter[Any], ParameterArg]


class MatchError(Exception):
    pass


class ImmutableStackIterator(Iterator[T]):
    def __init__(self, stack: Stack[T]) -> None:
        self.stack = stack

    def __next__(self) -> T:
        try:
            item, self.stack = self.stack.pop()
        except IndexError:
            raise StopIteration from None
        return item


@dataclasses.dataclass(slots=True, frozen=True)
class Stack(abc.ABC, Generic[T]):
    @staticmethod
    def from_sequence(items: Sequence[T]) -> Stack[T]:
        stack: Stack[T] = EmptyStack()

        for item in reversed(items):
            stack = stack.push(item)

        return stack

    @abc.abstractmethod
    def peek(self) -> T:
        ...

    @abc.abstractmethod
    def pop(self) -> tuple[T, Stack[T]]:
        ...

    def push(self, item: T) -> StackNode[T]:
        return StackNode(item, self)

    @abc.abstractmethod
    def __len__(self) -> int:
        ...

    @abc.abstractmethod
    def __bool__(self) -> bool:
        ...

    def __iter__(self) -> ImmutableStackIterator[T]:
        return ImmutableStackIterator(self)


@dataclasses.dataclass(slots=True, frozen=True)
class EmptyStack(Stack[T]):
    def peek(self) -> NoReturn:
        raise IndexError("Stack is empty")

    def pop(self) -> NoReturn:
        raise IndexError("Stack is empty")

    def __len__(self) -> int:
        return 0

    def __bool__(self) -> Literal[False]:
        return False

    def __repr__(self) -> str:
        return "Stack([])"


@dataclasses.dataclass(slots=True, frozen=True)
class StackNode(Stack[T]):
    head: T
    tail: Stack[T]

    def peek(self) -> T:
        return self.head

    def pop(self) -> tuple[T, Stack[T]]:
        return self.head, self.tail

    @functools.cache  # noqa: B019
    def __len__(self) -> int:
        return len(self.tail) + 1

    def __bool__(self) -> Literal[True]:
        return True

    def __repr__(self) -> str:
        return f"Stack({list(self)!r})"


@dataclasses.dataclass(slots=True)
class StackFrame:
    param: ParserParameter[Any]
    nargs: NArgs
    flag: str | None = None
    value: str | None = None
    has_value: bool = False


@dataclasses.dataclass(slots=True)
class State:
    args: Stack[str]
    positional: Stack[ArgumentParameter[Any]]
    parse_options: bool = True
    unused_args: Stack[str] = dataclasses.field(default_factory=EmptyStack)
    captured_args: Stack[CapturedArg] = dataclasses.field(default_factory=EmptyStack)
    stack: Stack[StackFrame] = dataclasses.field(default_factory=EmptyStack)

    def copy(self) -> State:
        return State(
            args=self.args,
            positional=self.positional,
            parse_options=self.parse_options,
            unused_args=self.unused_args,
            stack=self.stack,
            captured_args=self.captured_args,
        )

    def push_arg(self, arg: str) -> None:
        self.args = self.args.push(arg)

    def pop_arg(self) -> str:
        arg, self.args = self.args.pop()
        return arg

    def pop_positional(self) -> ArgumentParameter[Any]:
        param, self.positional = self.positional.pop()
        return param

    def push_unused_arg(self, arg: str) -> None:
        self.unused_args = self.unused_args.push(arg)

    def reset_unused_args(self) -> None:
        while self.unused_args:
            arg, self.unused_args = self.unused_args.pop()
            self.args = self.args.push(arg)

    def push_frame(self, frame: StackFrame) -> None:
        self.stack = self.stack.push(frame)

    def pop_frame(self) -> StackFrame:
        frame, self.stack = self.stack.pop()
        return frame

    def capture_arg(self, param: ParserParameter[Any], arg: ParameterArg) -> None:
        self.captured_args = self.captured_args.push((param, arg))

    def get_captured_args(self) -> dict[ParserParameter[Any], list[ParameterArg]]:
        args: dict[ParserParameter[Any], list[Any]] = {}

        for param, arg in self.captured_args:
            try:
                args[param].append(arg)
            except KeyError:
                args[param] = [arg]

        # Reverse the order of the values
        return {param: args[param][::-1] for param in args}

    def loss(self) -> tuple[int, int]:
        return len(self.positional), len(self.args)


def is_long_flag(arg: str) -> bool:
    return arg.startswith("--") and to_snake(arg[2:]).isidentifier()


def is_short_flag(arg: str) -> bool:
    return arg.startswith("-") and len(arg) == 2 and arg[1].isalpha()


def is_short_arg(arg: str) -> bool:
    return arg.startswith("-") and len(arg) > 1 and arg[1:].isalpha()


def is_long_arg(arg: str) -> bool:
    return is_long_flag(arg)


def is_positional_arg(arg: str) -> bool:
    return not (is_short_arg(arg) or is_long_arg(arg))


class Parser:
    def __init__(self, params: list[ParserParameter[Any]]) -> None:
        self.short_flags: dict[str, OptionParameter[Any]] = {}
        self.long_flags: dict[str, OptionParameter[Any]] = {}
        self.positional: list[ArgumentParameter[Any]] = []
        self.params: list[ParserParameter[Any]] = []
        self.add_params(params)

    def add_params(self, params: list[ParserParameter[Any]]) -> None:
        for param in params:
            self.add_param(param)

    def add_param(self, param: ParserParameter[Any]) -> None:
        if isinstance(param, ArgumentParameter):
            self.positional.append(param)

        elif isinstance(param, OptionParameter):
            for flag in itertools.chain(param.flags, param.secondary_flags):
                if is_short_flag(flag):
                    if flag in self.short_flags:
                        msg = f"Duplicate short flag: {flag}"
                        raise ValueError(msg)
                    self.short_flags[flag] = param

                elif is_long_flag(flag):
                    if flag in self.long_flags:
                        msg = f"Duplicate long flag: {flag}"
                        raise ValueError(msg)
                    self.long_flags[flag] = param

                else:
                    msg = f"Invalid flag: {flag}"
                    raise ValueError(msg)

        elif isinstance(param, DependsParameter):
            pass

        else:
            msg = f"Unsupported parameter type: {type(param)}"
            raise TypeError(msg)

        self.params.append(param)

    def parse_args(
        self,
        ctx: Context[Any],
        *,
        allow_interspersed_args: bool = True,
        raise_on_unexpected: bool = True,
    ) -> tuple[list[ParameterValue[Any]], list[str]]:
        state = self._parse_args(
            ctx,
            allow_interspersed_args=allow_interspersed_args,
        )

        captured_args = state.get_captured_args()
        values: list[ParameterValue[Any]] = []

        for param in self.params:
            value = param.process_args(ctx, captured_args.get(param, []))
            values.append(value)

        if (state.args or state.unused_args) and raise_on_unexpected:
            if not state.args:
                state.reset_unused_args()
            possibilities = self._get_possible_matches(ctx, state)
            raise UnexpectedParameter(state.args.peek(), possibilities=possibilities)

        state.reset_unused_args()
        return values, list(state.args)

    def _get_possible_matches(self, ctx: Context[Any], state: State) -> list[str]:
        flag, _, _ = state.args.peek().partition("=")
        if is_short_flag(flag):
            return []

        if is_long_flag(flag):
            return get_close_matches(flag, self.long_flags)

        if state.positional:
            return [state.positional.peek().get_metavar(ctx)]

        return []

    def _parse_args(
        self,
        ctx: Context[Any],
        *,
        allow_interspersed_args: bool = True,
    ) -> State:
        best_match: State = State(
            Stack.from_sequence(ctx.args),
            Stack.from_sequence(self.positional),
        )

        threads = [best_match]

        while threads:
            state = threads.pop()

            try:
                self._parse_thread(
                    ctx, threads, state, allow_interspersed_args=allow_interspersed_args
                )

            except MatchError:
                if state.loss() < best_match.loss():
                    best_match = state
                continue

            except UnexpectedParameter as error:
                state.push_arg(error.arg)

            return state

        return best_match

    def _parse_thread(
        self,
        ctx: Context[Any],
        threads: list[State],
        state: State,
        *,
        allow_interspersed_args: bool,
    ) -> None:
        while True:
            while state.stack:
                frame = state.pop_frame()
                self._parse_frame(ctx, threads, state, frame)

            if state.parse_options:
                self._parse_option(
                    ctx=ctx,
                    threads=threads,
                    state=state,
                    allow_interspersed_args=allow_interspersed_args,
                )

            elif state.positional:
                state.reset_unused_args()
                self._parse_positional(ctx=ctx, threads=threads, state=state)

            else:
                break

    def _push_frame(
        self,
        ctx: Context[Any],
        threads: list[State],
        state: State,
        param: ParserParameter[Any],
        nargs: NArgs | None,
        flag: str | None = None,
        value: str | None = None,
        has_value: bool = False,
    ) -> None:
        if nargs is None:
            return

        if isinstance(nargs, VariadicNArgs):
            if nargs.min_values == 0:
                if nargs.greedy:
                    # Try without in next thread
                    threads.append(state.copy())
                else:
                    # Move non-greedy variadics to the next thread
                    state = state.copy()
                    threads.append(state)

        state.push_frame(
            StackFrame(
                param=param,
                nargs=nargs,
                flag=flag,
                value=value,
                has_value=has_value,
            )
        )

    def _parse_frame(
        self,
        ctx: Context[Any],
        threads: list[State],
        state: State,
        frame: StackFrame,
    ) -> None:
        if isinstance(frame.nargs, IntNArgs):
            return self._parse_int_frame(ctx, threads, state, frame, frame.nargs)

        if isinstance(frame.nargs, VariadicNArgs):
            return self._parse_variadic_frame(ctx, threads, state, frame, frame.nargs)

        msg = f"Unsupported nargs: {frame.nargs!r}"
        raise TypeError(msg)

    def _parse_int_frame(
        self,
        ctx: Context[Any],
        threads: list[State],
        state: State,
        frame: StackFrame,
        nargs: IntNArgs,
    ) -> None:
        arg = self._get_int_frame_arg(ctx, state, frame, nargs)

        if not frame.param.matches(ctx, arg):
            raise MatchError()

        state.capture_arg(frame.param, arg)

        self._push_frame(
            ctx=ctx,
            threads=threads,
            state=state,
            param=frame.param,
            nargs=nargs.decrement(),
            flag=frame.flag,
        )

    def _get_int_frame_arg(
        self,
        ctx: Context[Any],
        state: State,
        frame: StackFrame,
        nargs: IntNArgs,
    ) -> ParameterArg:
        if frame.flag is not None:
            if nargs.value == 0 and frame.value is None:
                return OptionArg(flag=frame.flag, value=None)

            if nargs.value == 1 and frame.value is not None:
                return OptionArg(flag=frame.flag, value=frame.value)

        if frame.has_value:
            raise MatchError()

        try:
            value = state.pop_arg()
        except IndexError:
            raise MatchError() from None

        if frame.flag is not None:
            return OptionArg(flag=frame.flag, value=value)

        return PositionalArg(value=value)

    def _parse_variadic_frame(
        self,
        ctx: Context[Any],
        threads: list[State],
        state: State,
        frame: StackFrame,
        nargs: VariadicNArgs,
    ) -> None:
        child_frame = StackFrame(
            param=frame.param,
            nargs=nargs.value,
            flag=frame.flag,
        )

        self._parse_frame(ctx, threads, state, child_frame)

        self._push_frame(
            ctx=ctx,
            threads=threads,
            state=state,
            param=frame.param,
            nargs=nargs.decrement(),
            flag=frame.flag,
        )

    def _parse_option(
        self,
        ctx: Context[Any],
        threads: list[State],
        state: State,
        *,
        allow_interspersed_args: bool,
    ) -> None:
        if not state.args:
            state.parse_options = False
            return

        arg = state.pop_arg()
        if arg == "--":
            state.parse_options = False
            return

        value: str | None
        flag, has_value, value = arg.partition("=")

        if not has_value:
            value = None

        if is_short_arg(flag):
            self._parse_short(ctx, threads, state, arg, flag, value)

        elif is_long_arg(flag):
            self._parse_long(ctx, threads, state, arg, flag, value)

        elif allow_interspersed_args:
            state.push_unused_arg(arg)
        else:
            state.parse_options = False
            state.push_arg(arg)

    def _parse_short(
        self,
        ctx: Context[Any],
        threads: list[State],
        state: State,
        arg: str,
        flag: str,
        value: str | None,
    ) -> None:
        for index, short_name in enumerate(reversed(flag[1:])):
            short_flag = f"-{short_name}"

            try:
                param = self.short_flags[short_flag]
            except KeyError:
                raise UnexpectedParameter(arg) from None

            self._push_frame(
                ctx=ctx,
                threads=threads,
                state=state,
                param=param,
                nargs=param.nargs,
                flag=short_flag,
                value=value if index == 0 else None,
                has_value=index != 0 or value is not None,
            )

    def _parse_long(
        self,
        ctx: Context[Any],
        threads: list[State],
        state: State,
        arg: str,
        flag: str,
        value: str | None,
    ) -> None:
        try:
            param = self.long_flags[flag]
        except KeyError:
            raise UnexpectedParameter(arg) from None

        self._push_frame(
            ctx=ctx,
            threads=threads,
            state=state,
            param=param,
            nargs=param.nargs,
            flag=flag,
            value=value,
            has_value=value is not None,
        )

    def _parse_positional(
        self, ctx: Context[Any], threads: list[State], state: State
    ) -> None:
        param = state.pop_positional()
        self._push_frame(
            ctx=ctx,
            threads=threads,
            state=state,
            param=param,
            nargs=param.nargs,
        )


def parse_args(
    ctx: Context[Any],
    params: list[ParserParameter[Any]],
    *,
    allow_interspersed_args: bool = True,
    raise_on_unexpected: bool = True,
) -> tuple[list[ParameterValue[Any]], list[str]]:
    parser = Parser(params)
    return parser.parse_args(
        ctx,
        allow_interspersed_args=allow_interspersed_args,
        raise_on_unexpected=raise_on_unexpected,
    )



from __future__ import annotations

import collections.abc
import inspect
from collections.abc import Callable, Mapping, Sequence
from difflib import get_close_matches
from typing import (
    Any,
    ParamSpec,
    TypeVar,
    overload,
)

from tabb.base import BaseCommand
from tabb.callback import create_callback
from tabb.command import Command
from tabb.context import Context
from tabb.formatter import HelpFormatter
from tabb.parameter import ParameterGroup
from tabb.utils import to_kebab

P = ParamSpec("P")
T = TypeVar("T")


class Group(BaseCommand[T]):
    allow_interspersed_args: bool = False

    def __init__(
        self,
        callback: str | Callable[..., Any] | None = None,
        commands: Mapping[str, BaseCommand[T]] | Sequence[BaseCommand[T]] | None = None,
        *,
        add_help_option: bool = True,
        deprecated: bool = False,
        epilog: str | None = None,
        help: str | None = None,
        hidden: bool = False,
        import_dependencies: str | None = None,
        name: str | None = None,
        no_args_is_help: bool = True,
        options_metavar: str | None = "[OPTIONS]",
        package: str | None = None,
        short_help: str | None = None,
        subcommand_metavar: str = "COMMAND [ARGS]...",
    ) -> None:
        self._callback = create_callback(callback) if callback else None
        self.import_dependencies = import_dependencies
        self.package = package
        self.subcommand_metavar = subcommand_metavar
        self.commands: dict[str, BaseCommand[T]] = {}

        if commands is not None:
            self.add_commands(commands)

        super().__init__(
            add_help_option=add_help_option,
            deprecated=deprecated,
            epilog=epilog,
            help=help,
            hidden=hidden,
            name=name,
            no_args_is_help=no_args_is_help,
            options_metavar=options_metavar,
            short_help=short_help,
        )

    def get_name(self) -> str | None:
        name = super().get_name()
        if name:
            return name

        if self._callback:
            return to_kebab(self._callback.name)

        return None

    def get_help_text(self) -> str:
        help_text = super().get_help_text()
        if help_text:
            return help_text

        if self._callback:
            return inspect.getdoc(self._callback.fn) or ""

        return ""

    def collect_usage_pieces(self, ctx: Context[T]) -> list[str]:
        rv = super().collect_usage_pieces(ctx)
        rv.append(self.subcommand_metavar)
        return rv

    def format_options(self, ctx: Context[T], formatter: HelpFormatter) -> None:
        super().format_options(ctx, formatter)
        self.format_commands(ctx, formatter)

    def format_commands(self, ctx: Context[T], formatter: HelpFormatter) -> None:
        """Extra format methods for multi methods that adds all the commands
        after the options.
        """
        commands = {name: cmd for name, cmd in self.commands.items() if not cmd.hidden}
        if not commands:
            return

        limit = formatter.width - 6 - max(len(name) for name in commands)

        rows: list[tuple[str, str]] = []

        for name, cmd in commands.items():
            rows.append((name, cmd.get_short_help_text(limit)))

        with formatter.section("Commands"):
            formatter.write_dl(rows)

    def add_command(
        self, command: BaseCommand[Any], *, name: str | None = None
    ) -> None:
        if name is None:
            name = command.get_name()

        if name is None:
            raise ValueError("Command must have a name")

        self.commands[name] = command

    def add_commands(
        self, commands: Mapping[str, BaseCommand[Any]] | Sequence[BaseCommand[Any]]
    ) -> None:
        if isinstance(commands, collections.abc.Sequence):
            for command in commands:
                self.add_command(command)

        else:
            for name, command in commands.items():
                self.add_command(command, name=name)

    @overload
    def command(self, callback: Callable[P, T], /) -> Command[P, T]:
        ...

    @overload
    def command(
        self,
        name: str | None = None,
        *,
        add_help_option: bool = True,
        deprecated: bool = False,
        epilog: str | None = None,
        help: str | None = None,
        hidden: bool = False,
        import_dependencies: str | None = None,
        options_metavar: str | None = "[OPTIONS]",
        package: str | None = None,
        short_help: str | None = None,
    ) -> Callable[[Callable[P, T]], Command[P, T]]:
        ...

    def command(
        self,
        name: Callable[P, T] | str | None = None,
        **kwargs: Any,
    ) -> Command[P, T] | Callable[[Callable[P, T]], Command[P, T]]:
        if callable(name):
            command = Command(name)
            self.add_command(command)
            return command

        def decorator(callback: Callable[P, T]) -> Command[P, T]:
            if isinstance(callback, Command):
                raise TypeError("Attempted to convert a callback into a command twice.")

            command = Command(callback, name=name, **kwargs)
            self.add_command(command)
            return command

        return decorator

    @overload
    def group(self, callback: Callable[..., Any], /) -> Group[T]:
        ...

    @overload
    def group(
        self,
        name: str | None = None,
        *,
        add_help_option: bool = True,
        deprecated: bool = False,
        epilog: str | None = None,
        help: str | None = None,
        hidden: bool = False,
        import_dependencies: str | None = None,
        options_metavar: str | None = "[OPTIONS]",
        package: str | None = None,
        short_help: str | None = None,
    ) -> Callable[[Callable[..., Any]], Group[T]]:
        ...

    def group(
        self,
        name: Callable[..., Any] | str | None = None,
        **kwargs: Any,
    ) -> Group[T] | Callable[[Callable[..., Any]], Group[T]]:
        if callable(name):
            group: Group[T] = Group(name)
            self.add_command(group)
            return group

        def decorator(callback: Callable[..., Any]) -> Group[T]:
            if isinstance(callback, Group):
                raise TypeError("Attempted to convert a callback into a group twice.")

            group: Group[T] = Group(callback, name=name, **kwargs)
            self.add_command(group)
            return group

        return decorator

    def get_params(self, ctx: Context[T]) -> ParameterGroup:
        params = super().get_params(ctx)
        if self._callback is not None:
            params.add_params(self._callback.get_params(ctx))
        return params

    def invoke(self, ctx: Context[T]) -> T:
        args = self.parse_args(ctx, raise_on_unexpected=False)

        self.invoke_callbacks(ctx)

        if self._callback is not None:
            self._callback(ctx)

        try:
            command_name, args = args[0], args[1:]
        except IndexError:
            ctx.fail("Missing command.")

        try:
            command = self.commands[command_name]
        except KeyError:
            msg = f"Unknown command {command_name!r}."
            possibilities = get_close_matches(command_name, self.commands)

            if len(possibilities) == 1:
                msg += f" Did you mean {possibilities[0]!r}?"
            elif possibilities:
                options = ", ".join(sorted(possibilities))
                msg += f" (Possible options: {options})"

            ctx.fail(msg)


        sub_ctx = command.make_context(command_name, args, ctx.environ, ctx.config, ctx)

        with sub_ctx:
            return command.invoke(sub_ctx)



import re
import shutil
from collections.abc import Iterable, Iterator, Sequence
from contextlib import contextmanager
from textwrap import TextWrapper

_ansi_re = re.compile(r"\033\[[;?0-9]*[a-zA-Z]")


def strip_ansi(text: str) -> str:
    """Removes ANSI color codes from a string."""
    return _ansi_re.sub("", text)


def term_len(text: str) -> int:
    """Returns the length of a string as it would be displayed on the
    terminal.  This is done by stripping ANSI color codes first.  If
    you don't want that, use ``len(strip_ansi(text))`` instead.
    """
    return len(strip_ansi(text))


def measure_table(rows: Iterable[tuple[str, str]]) -> tuple[int, ...]:
    widths: dict[int, int] = {}

    for row in rows:
        for idx, col in enumerate(row):
            widths[idx] = max(widths.get(idx, 0), term_len(col))

    return tuple(y for x, y in sorted(widths.items()))


def iter_rows(
    rows: Iterable[tuple[str, str]], col_count: int
) -> Iterator[tuple[str, ...]]:
    for row in rows:
        yield row + ("",) * (col_count - len(row))


@contextmanager
def extra_indent(wrapper: TextWrapper, indent: str) -> Iterator[None]:
    old_initial_indent = wrapper.initial_indent
    old_subsequent_indent = wrapper.subsequent_indent
    wrapper.initial_indent += indent
    wrapper.subsequent_indent += indent

    try:
        yield
    finally:
        wrapper.initial_indent = old_initial_indent
        wrapper.subsequent_indent = old_subsequent_indent


def indent_only(wrapper: TextWrapper, text: str) -> str:
    rv = []

    for idx, line in enumerate(text.splitlines()):
        indent = wrapper.initial_indent

        if idx > 0:
            indent = wrapper.subsequent_indent

        rv.append(f"{indent}{line}")

    return "\n".join(rv)


def wrap_text(
    text: str,
    width: int = 78,
    initial_indent: str = "",
    subsequent_indent: str = "",
    preserve_paragraphs: bool = False,
) -> str:
    """A helper function that intelligently wraps text.  By default, it
    assumes that it operates on a single paragraph of text but if the
    `preserve_paragraphs` parameter is provided it will intelligently
    handle paragraphs (defined by two empty lines).

    If paragraphs are handled, a paragraph can be prefixed with an empty
    line containing the ``\\b`` character (``\\x08``) to indicate that
    no rewrapping should happen in that block.

    :param text: the text that should be rewrapped.
    :param width: the maximum width for the text.
    :param initial_indent: the initial indent that should be placed on the
                           first line as a string.
    :param subsequent_indent: the indent string that should be placed on
                              each consecutive line.
    :param preserve_paragraphs: if this flag is set then the wrapping will
                                intelligently handle paragraphs.
    """
    text = text.expandtabs()
    wrapper = TextWrapper(
        width,
        initial_indent=initial_indent,
        subsequent_indent=subsequent_indent,
        replace_whitespace=False,
    )
    if not preserve_paragraphs:
        return wrapper.fill(text)

    p: list[tuple[int, bool, str]] = []
    buf: list[str] = []
    indent = None

    def _flush_par() -> None:
        if not buf:
            return
        if buf[0].strip() == "\b":
            p.append((indent or 0, True, "\n".join(buf[1:])))
        else:
            p.append((indent or 0, False, " ".join(buf)))
        del buf[:]

    for line in text.splitlines():
        if not line:
            _flush_par()
            indent = None
        else:
            if indent is None:
                orig_len = term_len(line)
                line = line.lstrip()  # noqa: PLW2901
                indent = orig_len - term_len(line)
            buf.append(line)

    _flush_par()

    rv = []
    for indent, raw, text in p:
        with extra_indent(wrapper, " " * indent):
            if raw:
                rv.append(indent_only(wrapper, text))
            else:
                rv.append(wrapper.fill(text))

    return "\n\n".join(rv)


def make_default_short_help(help: str, max_length: int = 45) -> str:
    """Returns a condensed version of help string."""
    # Consider only the first paragraph.
    paragraph_end = help.find("\n\n")

    if paragraph_end != -1:
        help = help[:paragraph_end]  # noqa: A001

    # Collapse newlines, tabs, and spaces.
    words = help.split()

    if not words:
        return ""

    # The first paragraph started with a "no rewrap" marker, ignore it.
    if words[0] == "\b":
        words = words[1:]

    total_length = 0
    last_index = len(words) - 1

    for i, word in enumerate(words):
        total_length += len(word) + (i > 0)

        if total_length > max_length:  # too long, truncate
            break

        if word[-1] == ".":  # sentence end, truncate without "..."
            return " ".join(words[: i + 1])

        if total_length == max_length and i != last_index:
            break  # not at sentence end, truncate with "..."
    else:
        return " ".join(words)  # no truncation needed

    # Account for the length of the suffix.
    total_length += len("...")

    # remove words until the length is short enough
    while i > 0:
        total_length -= len(words[i]) + (i > 0)

        if total_length <= max_length:
            break

        i -= 1

    return " ".join(words[:i]) + "..."


class HelpFormatter:
    """This class helps with formatting text-based help pages.  It's
    usually just needed for very special internal cases, but it's also
    exposed so that developers can write their own fancy outputs.

    At present, it always writes into memory.

    :param indent_increment: the additional increment for each level.
    :param width: the width for the text.  This defaults to the terminal
                  width clamped to a maximum of 78.
    """

    def __init__(
        self,
        indent_increment: int = 2,
        width: int | None = None,
        max_width: int | None = None,
    ) -> None:
        self.indent_increment = indent_increment
        if max_width is None:
            max_width = 80
        if width is None:
            width = max(min(shutil.get_terminal_size().columns, max_width) - 2, 50)
        self.width = width
        self.current_indent = 0
        self.buffer: list[str] = []

    def write(self, string: str) -> None:
        """Writes a unicode string into the internal buffer."""
        self.buffer.append(string)

    def indent(self) -> None:
        """Increases the indentation."""
        self.current_indent += self.indent_increment

    def dedent(self) -> None:
        """Decreases the indentation."""
        self.current_indent -= self.indent_increment

    def write_usage(self, prog: str, args: str = "", prefix: str | None = None) -> None:
        """Writes a usage line into the buffer.

        :param prog: the program name.
        :param args: whitespace separated list of arguments.
        :param prefix: The prefix for the first line. Defaults to
            ``"Usage: "``.
        """
        if prefix is None:
            prefix = "Usage: "

        usage_prefix = f"{prefix:>{self.current_indent}}{prog} "
        text_width = self.width - self.current_indent

        if text_width >= (term_len(usage_prefix) + 20):
            # The arguments will fit to the right of the prefix.
            indent = " " * term_len(usage_prefix)
            self.write(
                wrap_text(
                    args,
                    text_width,
                    initial_indent=usage_prefix,
                    subsequent_indent=indent,
                )
            )
        else:
            # The prefix is too long, put the arguments on the next line.
            self.write(usage_prefix)
            self.write("\n")
            indent = " " * (max(self.current_indent, term_len(prefix)) + 4)
            self.write(
                wrap_text(
                    args, text_width, initial_indent=indent, subsequent_indent=indent
                )
            )

        self.write("\n")

    def write_heading(self, heading: str) -> None:
        """Writes a heading into the buffer."""
        self.write(f"{'':>{self.current_indent}}{heading}:\n")

    def write_paragraph(self) -> None:
        """Writes a paragraph into the buffer."""
        if self.buffer:
            self.write("\n")

    def write_text(self, text: str) -> None:
        """Writes re-indented text into the buffer.  This rewraps and
        preserves paragraphs.
        """
        indent = " " * self.current_indent
        self.write(
            wrap_text(
                text,
                self.width,
                initial_indent=indent,
                subsequent_indent=indent,
                preserve_paragraphs=True,
            )
        )
        self.write("\n")

    def write_dl(
        self,
        rows: Sequence[tuple[str, str]],
        col_max: int = 30,
        col_spacing: int = 2,
    ) -> None:
        """Writes a definition list into the buffer.  This is how options
        and commands are usually formatted.

        :param rows: a list of two item tuples for the terms and values.
        :param col_max: the maximum width of the first column.
        :param col_spacing: the number of spaces between the first and
                            second column.
        """
        rows = list(rows)
        widths = measure_table(rows)
        if len(widths) != 2:
            raise TypeError("Expected two columns for definition list")

        first_col = min(widths[0], col_max) + col_spacing

        for first, second in iter_rows(rows, len(widths)):
            self.write(f"{'':>{self.current_indent}}{first}")
            if not second:
                self.write("\n")
                continue
            if term_len(first) <= first_col - col_spacing:
                self.write(" " * (first_col - term_len(first)))
            else:
                self.write("\n")
                self.write(" " * (first_col + self.current_indent))

            text_width = max(self.width - first_col - 2, 10)
            wrapped_text = wrap_text(second, text_width, preserve_paragraphs=True)
            lines = wrapped_text.splitlines()

            if lines:
                self.write(f"{lines[0]}\n")

                for line in lines[1:]:
                    self.write(f"{'':>{first_col + self.current_indent}}{line}\n")
            else:
                self.write("\n")

    @contextmanager
    def section(self, name: str) -> Iterator[None]:
        """Helpful context manager that writes a paragraph, a heading,
        and the indents.

        :param name: the section name that is written as heading.
        """
        self.write_paragraph()
        self.write_heading(name)
        self.indent()
        try:
            yield
        finally:
            self.dedent()

    @contextmanager
    def indentation(self) -> Iterator[None]:
        """A context manager that increases the indentation."""
        self.indent()
        try:
            yield
        finally:
            self.dedent()

    def getvalue(self) -> str:
        """Returns the buffer contents."""
        return "".join(self.buffer)



from __future__ import annotations

import abc
import errno
import inspect
import os
import sys
from collections.abc import Callable, Mapping
from typing import (
    Any,
    Generic,
    NoReturn,
    ParamSpec,
    TypeVar,
)

from tabb.callback import Callback, create_callback
from tabb.context import Context
from tabb.exceptions import (
    Abort,
    Exit,
    Help,
    TabbError,
)
from tabb.formatter import HelpFormatter, make_default_short_help
from tabb.parameter import OptionParameter, ParameterGroup
from tabb.parameter.types import HelpFlag
from tabb.parser import parse_args
from tabb.utils import detect_program_name, pacify_flush

T = TypeVar("T")

CallbackParms = ParamSpec("CallbackParms")
CallbackResult = TypeVar("CallbackResult")


class BaseCommand(abc.ABC, Generic[T]):
    allow_interspersed_args: bool = True

    def __init__(
        self,
        *,
        add_help_option: bool,
        deprecated: bool,
        epilog: str | None,
        help: str | None,
        hidden: bool,
        name: str | None,
        no_args_is_help: bool,
        options_metavar: str | None,
        short_help: str | None,
    ) -> None:
        self.add_help_option = add_help_option
        self.deprecated = deprecated
        self.epilog = epilog
        self.help = help
        self.hidden = hidden
        self.name = name
        self.no_args_is_help = no_args_is_help
        self.options_metavar = options_metavar
        self.short_help = short_help
        self.callbacks: list[Callback[..., Any]] = []

    @abc.abstractmethod
    def invoke(self, ctx: Context[T]) -> T:
        raise NotImplementedError

    def make_context(
        self,
        name: str,
        args: list[str],
        environ: Mapping[str, str],
        config: Mapping[str, object],
        parent: Context[T] | None = None,
        *,
        auto_config_prefix: str | None = None,
        auto_envvar_prefix: str | None = None,
    ) -> Context[T]:
        return Context(
            name=name,
            args=args,
            environ=environ,
            config=config,
            command=self,
            parent=parent,
            auto_config_prefix=auto_config_prefix,
            auto_envvar_prefix=auto_envvar_prefix,
        )

    def get_name(self) -> str | None:
        return self.name

    def get_usage(self, ctx: Context[T]) -> str:
        """Formats the usage line into a string and returns it."""
        formatter = ctx.make_formatter()
        self.format_usage(ctx, formatter)
        return formatter.getvalue().rstrip("\n")

    def get_help(self, ctx: Context[T]) -> str:
        """Formats the help into a string and returns it."""
        formatter = ctx.make_formatter()
        self.format_help(ctx, formatter)
        return formatter.getvalue().rstrip("\n")

    def get_help_text(self) -> str:
        if self.help:
            return inspect.cleandoc(self.help)
        return ""

    def get_short_help_text(self, limit: int = 45) -> str:
        if self.short_help:
            text = self.short_help
        else:
            text = make_default_short_help(self.get_help_text(), limit)

        if self.deprecated:
            text = f"(Deprecated) {text}"

        return text.strip()

    def get_help_option(self, ctx: Context[T]) -> OptionParameter[NoReturn] | None:
        """Returns the help option object."""
        if not self.add_help_option:
            return None

        help_flags = self.get_help_option_names(ctx)
        if not help_flags:
            return None

        return OptionParameter(
            name="help",
            kind=OptionParameter.POSITIONAL,
            flags=help_flags,
            type=HelpFlag(),
            help="Show this message and exit.",
            required=False,
        )

    def get_help_option_names(self, ctx: Context[T]) -> list[str]:
        """Returns the names for the help option."""
        return ["--help", "-h"]

    def format_usage(self, ctx: Context[T], formatter: HelpFormatter) -> None:
        """Writes the usage line into the formatter if it exists."""
        pieces = self.collect_usage_pieces(ctx)
        formatter.write_usage(ctx.command_path, " ".join(pieces))

    def format_help(self, ctx: Context[T], formatter: HelpFormatter) -> None:
        """Writes the help into the formatter if it exists.

        This calls the following methods:

        -   :meth:`format_usage`
        -   :meth:`format_help_text`
        -   :meth:`format_options`
        -   :meth:`format_epilog`
        """
        self.format_usage(ctx, formatter)
        self.format_help_text(ctx, formatter)
        self.format_options(ctx, formatter)
        self.format_epilog(ctx, formatter)

    def format_help_text(self, ctx: Context[T], formatter: HelpFormatter) -> None:
        """Writes the help text to the formatter if it exists."""
        text = self.get_help_text()

        if text:
            # truncate the help text to the first form feed
            text = inspect.cleandoc(text).partition("\f")[0]

        if self.deprecated:
            text = f"(Deprecated) {text}"

        if text:
            formatter.write_paragraph()

            with formatter.indentation():
                formatter.write_text(text)

    def format_options(self, ctx: Context[T], formatter: HelpFormatter) -> None:
        """Writes all the options into the formatter if they exist."""
        opts = []
        for param in self.get_params(ctx).get_parser_params():
            rv = param.get_help_record(ctx)
            if rv is not None:
                opts.append(rv)

        if opts:
            with formatter.section("Options"):
                formatter.write_dl(opts)

    def format_epilog(self, ctx: Context[T], formatter: HelpFormatter) -> None:
        """Writes the epilog into the formatter if it exists."""
        if self.epilog:
            epilog = inspect.cleandoc(self.epilog)
            formatter.write_paragraph()

            with formatter.indentation():
                formatter.write_text(epilog)

    def collect_usage_pieces(self, ctx: Context[T]) -> list[str]:
        """Returns all the pieces that go into the usage line and returns
        it as a list of strings.
        """
        usage_pieces = [self.options_metavar] if self.options_metavar else []

        for param in self.get_params(ctx).get_parser_params():
            usage_pieces.extend(param.get_usage_pieces(ctx))

        return usage_pieces

    def callback(
        self, callback: Callable[CallbackParms, CallbackResult], /
    ) -> Callable[CallbackParms, CallbackResult]:
        self.add_callback(callback)
        return callback

    def add_callback(
        self, fn: Callable[..., Any] | str, *, import_dependencies: str | None = None
    ) -> None:
        self.callbacks.append(
            create_callback(fn, import_dependencies=import_dependencies)
        )

    def get_params(self, ctx: Context[T]) -> ParameterGroup:
        params = ParameterGroup(ctx)

        if (help_option := self.get_help_option(ctx)) is not None:
            params.add_param(help_option)

        for callback in self.callbacks:
            params.add_params(callback.get_params(ctx))

        return params

    def parse_args(
        self, ctx: Context[T], *, raise_on_unexpected: bool = True
    ) -> list[str]:
        if not ctx.args and self.no_args_is_help:
            raise Help(ctx=ctx)

        params = self.get_params(ctx)

        values, remaining_args = parse_args(
            ctx,
            params.get_parser_params(),
            allow_interspersed_args=self.allow_interspersed_args,
            raise_on_unexpected=raise_on_unexpected,
        )

        for value in values:
            ctx.values[value.param] = value

        for param in params.get_depends_params():
            ctx.values[param] = param.resolve(ctx)

        return remaining_args

    def invoke_callbacks(self, ctx: Context[T]) -> None:
        for callback in self.callbacks:
            callback(ctx)

    def run(
        self,
        args: list[str] | None = None,
        *,
        prog_name: str | None = None,
        environ: Mapping[str, str] | None = None,
        config: Mapping[str, object] | None = None,
        auto_config_prefix: str | None = None,
        auto_envvar_prefix: str | None = None,
    ) -> T:
        if args is None:
            args = sys.argv[1:]
        else:
            args = list(args)

        if prog_name is None:
            prog_name = detect_program_name()

        if environ is None:
            environ = os.environ

        if config is None:
            config = {}

        ctx = self.make_context(
            prog_name,
            args,
            environ,
            config,
            auto_config_prefix=auto_config_prefix,
            auto_envvar_prefix=auto_envvar_prefix,
        )

        with ctx:
            return self.invoke(ctx)

    def main(
        self,
        args: list[str] | None = None,
        *,
        prog_name: str | None = None,
        environ: Mapping[str, str] | None = None,
        config: Mapping[str, object] | None = None,
        auto_config_prefix: str | None = None,
        auto_envvar_prefix: str | None = None,
    ) -> NoReturn:
        try:
            try:
                self.run(
                    args=args,
                    prog_name=prog_name,
                    environ=environ,
                    config=config,
                    auto_config_prefix=auto_config_prefix,
                    auto_envvar_prefix=auto_envvar_prefix,
                )
                raise Exit()

            except TabbError as error:
                error.show()
                sys.exit(error.exit_code)

            except (EOFError, KeyboardInterrupt):
                print(file=sys.stderr)
                raise Abort() from None

            except OSError as error:
                if error.errno == errno.EPIPE:
                    sys.stdout = pacify_flush(sys.stdout)
                    sys.stderr = pacify_flush(sys.stderr)
                    sys.exit(1)
                else:
                    raise

        except Exit as error:
            sys.exit(error.exit_code)

        except Abort:
            print("Aborted!", file=sys.stderr)
            sys.exit(1)

    def __call__(
        self,
        args: list[str] | None = None,
        *,
        prog_name: str | None = None,
        environ: Mapping[str, str] | None = None,
        config: Mapping[str, object] | None = None,
        auto_config_prefix: str | None = None,
        auto_envvar_prefix: str | None = None,
    ) -> NoReturn:
        self.main(
            args,
            prog_name=prog_name,
            environ=environ,
            config=config,
            auto_config_prefix=auto_config_prefix,
            auto_envvar_prefix=auto_envvar_prefix,
        )



from __future__ import annotations

import abc
import dataclasses
from itertools import repeat
from typing import Literal

NArgsLiteral = int | Literal["?", "*", "+", "??", "*?", "+?"]


@dataclasses.dataclass(slots=True, frozen=True)
class NArgs(abc.ABC):
    @staticmethod
    def from_literal(value: NArgsLiteral) -> NArgs:
        if isinstance(value, int):
            return IntNArgs(value)

        if value == "?":
            return VariadicNArgs(IntNArgs(1), min_values=0, max_values=1, greedy=True)

        if value == "*":
            return VariadicNArgs(
                IntNArgs(1), min_values=0, max_values=None, greedy=True
            )

        if value == "+":
            return VariadicNArgs(
                IntNArgs(1), min_values=1, max_values=None, greedy=True
            )

        if value == "??":
            return VariadicNArgs(IntNArgs(1), min_values=0, max_values=1, greedy=False)

        if value == "*?":
            return VariadicNArgs(
                IntNArgs(1), min_values=0, max_values=None, greedy=False
            )

        if value == "+?":
            return VariadicNArgs(
                IntNArgs(1), min_values=1, max_values=None, greedy=False
            )

        msg = f"Invalid nargs literal: {value!r}"
        raise ValueError(msg)

    def as_optional(self) -> NArgs:
        return VariadicNArgs(self, min_values=0, max_values=1, greedy=True)

    @abc.abstractmethod
    def decrement(self) -> NArgs | None:
        ...

    @abc.abstractmethod
    def format_metavar(self, metavar: str) -> str:
        ...

    @abc.abstractmethod
    def __str__(self) -> str:
        ...


@dataclasses.dataclass(slots=True, frozen=True)
class IntNArgs(NArgs):
    value: int

    def __post_init__(self) -> None:
        if self.value < 0:
            raise ValueError("nargs cannot be negative.")

    def decrement(self) -> IntNArgs | None:
        if self.value <= 1:
            return None
        return IntNArgs(self.value - 1)

    def format_metavar(self, metavar: str) -> str:
        if self.value == 0:
            return ""

        if self.value == 1:
            return metavar

        if self.value <= 16 // len(metavar):
            return " ".join(repeat(metavar, self.value))

        return f"{metavar}{{{self.value}}}"

    def __str__(self) -> str:
        return f"{{{self.value}}}"


@dataclasses.dataclass(slots=True, frozen=True)
class VariadicNArgs(NArgs):
    value: NArgs
    min_values: int
    max_values: int | None
    greedy: bool = True

    def __post_init__(self) -> None:
        if self.min_values < 0:
            raise ValueError("min_values cannot be negative.")

        if self.max_values is None:
            return

        if self.min_values > self.max_values:
            raise ValueError("min_values cannot be greater than max_values.")

    def as_optional(self) -> NArgs:
        if self.min_values == 0:
            return self
        return super(VariadicNArgs, self).as_optional()  # noqa: UP008

    def decrement(self) -> VariadicNArgs | None:
        if self.max_values == 1:
            return None

        return VariadicNArgs(
            value=self.value,
            min_values=self.min_values and self.min_values - 1,
            max_values=self.max_values and self.max_values - 1,
            greedy=self.greedy,
        )

    def _format_range(self) -> str:
        if self.min_values == self.max_values:
            return f"{{{self.min_values}}}"

        max_value = "" if self.max_values is None else self.max_values
        greedy = "" if self.greedy else "?"

        return f"{{{self.min_values},{max_value}}}{greedy}"

    def _format_modifier(self) -> str:
        greedy = "" if self.greedy else "?"

        if self.min_values == 0 and self.max_values == 1:
            return f"?{greedy}"

        if self.min_values == 0 and self.max_values is None:
            return f"*{greedy}"

        if self.min_values == 1 and self.max_values is None:
            return f"+{greedy}"

        return self._format_range()

    def format_metavar(self, metavar: str) -> str:
        result = self.value.format_metavar(metavar)
        if not result:
            return ""

        if self.min_values == 0 and self.max_values == 0:
            return ""

        if self.min_values == 1 and self.max_values == 1:
            return result

        if result == metavar and self.min_values == self.max_values:
            return IntNArgs(self.min_values).format_metavar(metavar)

        if result != metavar:
            result = f"({result})"

        return f"{result}{self._format_modifier()}"

    def __str__(self) -> str:
        result = str(self.value)
        if result == "{0}":
            return "{0}"

        if self.min_values == 0 and self.max_values == 0:
            return "{0}"

        if self.min_values == 1 and self.max_values == 1:
            return result

        if result == "{1}":
            return self._format_range()

        return f"({result}){self._format_modifier()}"



from __future__ import annotations

from collections.abc import Callable, Mapping
from contextlib import AbstractContextManager, ExitStack
from threading import local
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    NoReturn,
    ParamSpec,
    TypeVar,
    cast,
    overload,
)

from tabb.config import Config
from tabb.exceptions import Abort, Exit, UsageError
from tabb.formatter import HelpFormatter
from tabb.utils import to_kebab, to_snake

if TYPE_CHECKING:
    from tabb.base import BaseCommand
    from tabb.parameter import Parameter, ParameterValue


T = TypeVar("T")
V = TypeVar("V")
P = ParamSpec("P")


class DependencyCache:
    def __init__(self) -> None:
        self.callbacks: dict[Callable[..., Any], Callable[[Context[Any]], Any]] = {}
        self.values: dict[Callable[[Context[Any]], Any], Any] = {}


_local = local()


@overload
def get_current_context(silent: Literal[False] = False) -> Context[Any]:
    ...


@overload
def get_current_context(silent: bool = ...) -> Context[Any] | None:
    ...


def get_current_context(silent: bool = False) -> Context[Any] | None:
    try:
        return cast("Context", _local.stack[-1])
    except (AttributeError, IndexError) as error:
        if not silent:
            raise RuntimeError("There is no active click context.") from error

    return None


def push_context(ctx: Context[Any]) -> None:
    """Pushes a new context to the current stack."""
    _local.__dict__.setdefault("stack", []).append(ctx)


def pop_context() -> None:
    """Removes the top level from the stack."""
    _local.stack.pop()


class Context(Generic[T]):
    def __init__(
        self,
        name: str,
        args: list[str],
        environ: Mapping[str, str],
        config: Mapping[str, object],
        command: BaseCommand[T],
        parent: Context[Any] | None = None,
        auto_config_prefix: str | None = None,
        auto_envvar_prefix: str | None = None,
    ) -> None:
        self.name = name
        self.args = args
        self.environ = environ
        self.config = Config(config)
        self.command = command
        self.parent = parent
        self.values: dict[Parameter[Any], ParameterValue[Any]] = {}

        if auto_config_prefix is None:
            if parent is not None and parent.auto_config_prefix is not None:
                auto_config_prefix = f"{parent.auto_config_prefix}.{self.name.lower()}"
        else:
            auto_config_prefix = auto_config_prefix.lower()

        if auto_config_prefix is not None:
            auto_config_prefix = to_kebab(auto_config_prefix)

        self.auto_config_prefix: str | None = auto_config_prefix

        if auto_envvar_prefix is None:
            if parent is not None and parent.auto_envvar_prefix is not None:
                auto_envvar_prefix = f"{parent.auto_envvar_prefix}_{self.name.upper()}"
        else:
            auto_envvar_prefix = auto_envvar_prefix.upper()

        if auto_envvar_prefix is not None:
            auto_envvar_prefix = to_snake(auto_envvar_prefix)

        self.auto_envvar_prefix: str | None = auto_envvar_prefix

        if parent is None:
            self.dependencies = DependencyCache()
        else:
            self.dependencies = parent.dependencies

        self._depth = 0
        self._exit_stack = ExitStack()

    @property
    def command_path(self) -> str:
        if self.parent is None:
            return self.name

        path: list[str] = [self.parent.command_path]

        for param in self.parent.command.get_params(self).get_parser_params():
            path.extend(param.get_usage_pieces(self))

        path.append(self.name)
        return " ".join(path)

    def make_formatter(self) -> HelpFormatter:
        return HelpFormatter()

    def get_help(self) -> str:
        return self.command.get_help(self)

    def get_usage(self) -> str:
        return self.command.get_usage(self)

    def abort(self) -> NoReturn:
        raise Abort()

    def exit(self, code: int = 0) -> NoReturn:
        raise Exit(code)

    def fail(self, message: str) -> NoReturn:
        raise UsageError(message, self)

    def with_resource(self, context_manager: AbstractContextManager[V]) -> V:
        return self._exit_stack.enter_context(context_manager)

    def call_on_close(self, fn: Callable[P, V]) -> Callable[P, V]:
        return self._exit_stack.callback(fn)

    def close(self) -> None:
        self._exit_stack.close()
        # In case the context is reused, create a new exit stack.
        self._exit_stack = ExitStack()

    def __enter__(self) -> Context[T]:
        self._depth += 1
        push_context(self)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        self._depth = min(self._depth - 1, 0)
        if self._depth == 0:
            self.close()
        pop_context()



from tabb.__about__ import __version__
from tabb.command import Command
from tabb.config import Config
from tabb.context import Context
from tabb.decorators import command, group
from tabb.exceptions import (
    Abort,
    BadParameter,
    Exit,
    Help,
    MissingParameter,
    TabbError,
    UnexpectedParameter,
    UsageError,
)
from tabb.group import Group
from tabb.missing import MISSING, is_missing
from tabb.nargs import NArgs, NArgsLiteral
from tabb.parameter import (
    Bool,
    Counter,
    Dict,
    File,
    Flag,
    Float,
    Int,
    List,
    Optional,
    Path,
    String,
    Tuple,
)
from tabb.types import (
    Argument,
    Choices,
    Depends,
    Greedy,
    Length,
    Matches,
    Option,
    Range,
    Secret,
    Validate,
)

__all__ = [
    "__version__",
    "Abort",
    "Argument",
    "BadParameter",
    "Bool",
    "Choices",
    "command",
    "Command",
    "Config",
    "Context",
    "Counter",
    "Depends",
    "Dict",
    "Exit",
    "File",
    "Flag",
    "Float",
    "Greedy",
    "group",
    "Group",
    "Help",
    "Int",
    "is_missing",
    "Length",
    "List",
    "Matches",
    "MISSING",
    "MissingParameter",
    "NArgs",
    "NArgsLiteral",
    "Option",
    "Optional",
    "Path",
    "Range",
    "Secret",
    "String",
    "TabbError",
    "Tuple",
    "UnexpectedParameter",
    "UsageError",
    "Validate",
]



from tabb.parameter.base import Parameter, ParameterValue
from tabb.parameter.depends import DependsParameter
from tabb.parameter.group import ParameterGroup
from tabb.parameter.parser import (
    ArgumentParameter,
    OptionParameter,
    ParserParameter,
)
from tabb.parameter.resolve import resolve_params
from tabb.parameter.types import (
    Bool,
    Counter,
    Dict,
    File,
    Flag,
    Float,
    Int,
    List,
    Optional,
    ParameterType,
    Path,
    Scalar,
    SequenceType,
    String,
    Tuple,
)

__all__ = [
    "ArgumentParameter",
    "Bool",
    "Counter",
    "DependsParameter",
    "Dict",
    "File",
    "Flag",
    "Float",
    "Int",
    "List",
    "Optional",
    "OptionParameter",
    "Parameter",
    "ParameterGroup",
    "ParameterType",
    "ParameterValue",
    "ParserParameter",
    "Path",
    "resolve_params",
    "Scalar",
    "SequenceType",
    "String",
    "Tuple",
]



from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from tabb.parameter.depends import DependsParameter
from tabb.parameter.parser import ParserParameter

if TYPE_CHECKING:
    from tabb.context import Context
    from tabb.parameter.base import Parameter


class ParameterGroup:
    def __init__(
        self,
        ctx: Context[Any],
        params: Iterable[Parameter[Any]] = (),
    ) -> None:
        self.ctx = ctx
        self._params: dict[Parameter[Any], None] = {}
        self._parser_params: list[ParserParameter[Any]] = []
        self._depends_params: list[DependsParameter[Any]] = []

    def add_params(self, params: Iterable[Parameter[Any]]) -> None:
        for param in params:
            self.add_param(param)

    def add_param(self, param: Parameter[Any]) -> None:
        if param in self._params:
            return

        if isinstance(param, ParserParameter):
            self._parser_params.append(param)

        elif isinstance(param, DependsParameter):
            self.add_params(param.get_sub_params(self.ctx))
            self._depends_params.append(param)

        self._params[param] = None

    def get_params(self) -> list[Parameter[Any]]:
        return list(self._params)

    def get_parser_params(self) -> list[ParserParameter[Any]]:
        return list(self._parser_params)

    def get_depends_params(self) -> list[DependsParameter[Any]]:
        return list(self._depends_params)



from __future__ import annotations

import abc
import dataclasses
import enum
from typing import Any, Generic, TypeVar

from tabb.context import Context
from tabb.missing import _MISSING_TYPE

ValueType = TypeVar("ValueType")


class ParameterKind(enum.Enum):
    POSITIONAL = enum.auto()
    KEYWORD = enum.auto()
    POSITIONAL_OR_KEYWORD = enum.auto()
    VAR_POSITIONAL = enum.auto()
    VAR_KEYWORD = enum.auto()


class ParameterSource(enum.Enum):
    COMMANDLINE = enum.auto()
    ENVIRONMENT = enum.auto()
    CONFIG = enum.auto()
    DEFAULT = enum.auto()
    PROMPT = enum.auto()
    DEPENDS = enum.auto()


@dataclasses.dataclass(frozen=True, slots=True)
class ParameterValue(Generic[ValueType]):
    value: ValueType | _MISSING_TYPE
    source: ParameterSource
    param: Parameter[ValueType]


class Parameter(abc.ABC, Generic[ValueType]):
    POSITIONAL = ParameterKind.POSITIONAL
    KEYWORD = ParameterKind.KEYWORD
    POSITIONAL_OR_KEYWORD = ParameterKind.POSITIONAL_OR_KEYWORD
    VAR_POSITIONAL = ParameterKind.VAR_POSITIONAL
    VAR_KEYWORD = ParameterKind.VAR_KEYWORD

    def __init__(
        self,
        name: str,
        kind: ParameterKind,
    ) -> None:
        self.name = name
        self.kind = kind

    def get_param_type(self) -> str:
        return "parameter"

    def get_usage_pieces(self, ctx: Context[Any] | None = None) -> list[str]:
        return []

    def get_help_record(
        self, ctx: Context[Any] | None = None
    ) -> tuple[str, str] | None:
        return None



from __future__ import annotations

import inspect
import pathlib
import typing
from collections.abc import Callable
from typing import Any

from tabb.context import Context
from tabb.missing import _MISSING_TYPE, MISSING
from tabb.nargs import NArgs, NArgsLiteral
from tabb.parameter.base import Parameter, ParameterKind
from tabb.parameter.depends import DependsParameter
from tabb.parameter.parser import ArgumentParameter, OptionParameter
from tabb.parameter.types import (
    AnyType,
    Bool,
    ChoicesValidator,
    Dict,
    Flag,
    GreedyAdaptor,
    LengthValidator,
    List,
    Optional,
    ParameterType,
    Path,
    RangeValidator,
    Scalar,
    Tuple,
    Validator,
)
from tabb.parameter.utils import (
    get_annotations,
    get_base_type,
    get_list_child_type,
    get_literal_child_type,
    get_optional_child_type,
    get_tuple_child_type,
)
from tabb.types import (
    Argument,
    Choices,
    Depends,
    Greedy,
    Length,
    Matches,
    Option,
    Range,
    Validate,
)
from tabb.utils import to_kebab


def get_context(ctx: Context[Any]) -> Context[Any]:
    return ctx


def resolve_params(
    fn: Callable[..., Any], *, localns: dict[str, Any] | None = None
) -> list[Parameter[Any]]:
    params: list[Parameter[Any]] = []
    signature = inspect.signature(fn)
    type_hints = typing.get_type_hints(fn, localns=localns, include_extras=True)

    for name, param_signature in signature.parameters.items():
        try:
            params.append(_resolve_parameter(name, param_signature, type_hints[name]))
        except TypeError as error:
            msg = f"Invalid type for parameter {name!r}. {error}"
            raise TypeError(msg) from error

    return params


def _resolve_parameter(
    name: str, signature: inspect.Parameter, type_hint: object
) -> Parameter[Any]:
    param: Parameter[Any]

    # Arguments with a default of None are automatically made optional
    is_optional, type_hint = get_optional_child_type(type_hint)
    type_hint, annotations = get_annotations(type_hint)
    options = _resolve_options(signature, type_hint, annotations)

    if isinstance(options, Argument):
        param = _resolve_argument(
            name, signature, type_hint, annotations, options, is_optional
        )

    elif isinstance(options, Option):
        param = _resolve_option(
            name, signature, type_hint, annotations, options, is_optional
        )

    elif isinstance(options, Depends):
        return _get_depends(name, signature, options)

    else:
        msg = f"Unexpected options type: {options!r}."
        raise TypeError(msg)

    if options.default is not MISSING:
        param.default = options.default

    elif options.default_factory is not MISSING:
        param.default_factory = options.default_factory

    elif signature.default is not signature.empty:
        param.default = signature.default

    if isinstance(options.config, str):
        param.config = [options.config]
    elif not isinstance(options.config, _MISSING_TYPE):
        param.config = list(options.config)

    if isinstance(options.envvar, str):
        param.envvar = [options.envvar]
    elif not isinstance(options.envvar, _MISSING_TYPE):
        param.envvar = list(options.envvar)

    if not isinstance(options.help, _MISSING_TYPE):
        param.help = options.help

    if not isinstance(options.hidden, _MISSING_TYPE):
        param.hidden = options.hidden

    if not isinstance(options.metavar, _MISSING_TYPE):
        param.metavar = options.metavar

    return param


def _resolve_argument(
    name: str,
    signature: inspect.Parameter,
    type_hint: object,
    annotations: list[object],
    options: Argument[Any],
    is_optional: bool,
) -> ArgumentParameter[Any]:
    parameter_type = _get_argument_type(type_hint, options, is_optional)
    parameter_type, kind = _resolve_parameter_kind(signature, parameter_type, nargs="*")
    parameter_type = _resolve_validators(type_hint, annotations, parameter_type)
    return ArgumentParameter(name, kind, parameter_type)


def _resolve_option(
    name: str,
    signature: inspect.Parameter,
    type_hint: object,
    annotations: list[object],
    options: Option[Any],
    is_optional: bool,
) -> OptionParameter[Any]:
    parameter_type = _resolve_option_type(type_hint, options, is_optional)
    parameter_type, kind = _resolve_parameter_kind(signature, parameter_type)
    parameter_type = _resolve_validators(type_hint, annotations, parameter_type)
    flags = list(options.flags) if options.flags else [f"--{to_kebab(name)}"]
    option = OptionParameter(name, kind, parameter_type, flags)

    if not isinstance(options.show_config, _MISSING_TYPE):
        option.show_config = options.show_config

    if not isinstance(options.show_default, _MISSING_TYPE):
        option.show_default = options.show_default

    if not isinstance(options.show_envvar, _MISSING_TYPE):
        option.show_envvar = options.show_envvar

    return option


def _get_depends(
    name: str,
    signature: inspect.Parameter,
    options: Depends[Any],
) -> DependsParameter[Any]:
    _, kind = _resolve_parameter_kind(signature, AnyType())
    return DependsParameter(name, kind, options.dependency, use_cache=options.use_cache)


def _get_argument_type(
    type_hint: object,
    options: Argument[Any],
    is_optional: bool = False,
) -> ParameterType[Any]:
    if not isinstance(options.type, _MISSING_TYPE):
        return options.type

    # Resolve optional types
    child_is_optional, type_hint = get_optional_child_type(type_hint)
    if is_optional or child_is_optional:
        item_type = _get_argument_type(type_hint, options)
        return Optional(item_type)

    # Resolve list types
    is_list, child_type = get_list_child_type(type_hint)
    if is_list:
        item_type = _get_argument_type(child_type, options)
        return List(item_type, nargs="*")

    # Resolve tuple types
    is_tuple, child_type = get_tuple_child_type(type_hint)
    if is_tuple:
        item_type = _get_argument_type(child_type, options)
        return Tuple(item_type, nargs="*")

    # Resolve Literal types
    is_literal, literal_type = get_literal_child_type(type_hint)
    if is_literal:
        choices = list(typing.get_args(type_hint))
        child_type = _get_argument_type(literal_type, options)
        return ChoicesValidator(child_type, choices)

    # Handle bool aruments
    if type_hint is bool:
        return Bool()

    if (result := _resolve_special_types(type_hint)) is not None:
        return result

    if callable(type_hint):
        return Scalar(type_hint)

    msg = f"Unexpted type: {type_hint!r}."
    raise TypeError(msg)


def _resolve_option_type(
    type_hint: object,
    options: Option[Any],
    is_optional: bool = False,
) -> ParameterType[Any]:
    if not isinstance(options.type, _MISSING_TYPE):
        return options.type

    # Resolve optional types
    child_is_optional, type_hint = get_optional_child_type(type_hint)
    if is_optional or child_is_optional:
        item_type = _resolve_option_type(type_hint, options)
        return Optional(item_type)

    # Resolve list types
    is_list, child_type = get_list_child_type(type_hint)
    if is_list:
        item_type = _resolve_option_type(child_type, options)
        return List(item_type)

    # Resolve tuple types
    is_tuple, child_type = get_tuple_child_type(type_hint)
    if is_tuple:
        item_type = _resolve_option_type(child_type, options)
        return Tuple(item_type)

    # Resolve Literal types
    is_literal, literal_type = get_literal_child_type(type_hint)
    if is_literal:
        choices = list(typing.get_args(type_hint))
        child_type = _resolve_option_type(literal_type, options)
        return ChoicesValidator(child_type, choices)

    # Bool options become flags
    if type_hint is bool:
        return Flag()

    if (result := _resolve_special_types(type_hint)) is not None:
        return result

    if callable(type_hint):
        return Scalar(type_hint)

    msg = f"Unexpted type: {type_hint!r}."
    raise TypeError(msg)


def _resolve_parameter_kind(
    signature: inspect.Parameter,
    parameter_type: ParameterType[Any],
    nargs: NArgs | NArgsLiteral | None = None,
) -> tuple[ParameterType[Any], ParameterKind]:
    if signature.kind == signature.VAR_POSITIONAL:
        parameter_type = List(parameter_type, nargs=nargs)
        return parameter_type, ParameterKind.VAR_POSITIONAL

    if signature.kind == signature.VAR_KEYWORD:
        parameter_type = Dict(parameter_type, nargs=nargs)
        return parameter_type, ParameterKind.VAR_KEYWORD

    if signature.kind == signature.KEYWORD_ONLY:
        return parameter_type, ParameterKind.KEYWORD

    if signature.kind == signature.POSITIONAL_ONLY:
        return parameter_type, ParameterKind.POSITIONAL

    if signature.kind == signature.POSITIONAL_OR_KEYWORD:
        return parameter_type, ParameterKind.POSITIONAL_OR_KEYWORD

    msg = f"Unexpted parameter kind: {signature.kind!r}."
    raise TypeError(msg)


def _resolve_special_types(
    type_hint: object,
) -> ParameterType[Any] | None:
    base_type = get_base_type(type_hint)

    if base_type is pathlib.Path:
        return Path()

    return None


def _resolve_validators(
    type_hint: object,
    annotations: list[object],
    parameter_type: ParameterType[Any],
) -> ParameterType[Any]:
    for annotation in annotations:
        if isinstance(annotation, Length):
            parameter_type = LengthValidator(
                type=parameter_type,
                min_length=annotation.min,
                max_length=annotation.max,
            )

        elif isinstance(annotation, Range):
            cast = get_base_type(type_hint)
            if cast not in (int, float):
                msg = f"Cannot use min/max with type {cast!r}."
                raise TypeError(msg)

            parameter_type = RangeValidator(
                type=parameter_type,
                cast=cast,
                min=annotation.min,
                max=annotation.max,
                min_open=annotation.min_open,
                max_open=annotation.max_open,
                clamp=annotation.clamp,
            )

        elif isinstance(annotation, Choices):
            parameter_type = ChoicesValidator(
                type=parameter_type,
                choices=annotation.values,
            )

        elif isinstance(annotation, Matches):
            # FIXME: This is not implemented yet
            pass

        elif isinstance(annotation, Greedy):
            parameter_type = GreedyAdaptor(
                type=parameter_type, greedy=annotation.is_greedy
            )

        elif isinstance(annotation, Validate):
            parameter_type = Validator(parameter_type, annotation.validator)

    return parameter_type


def _resolve_options(
    signature: inspect.Parameter, type_hint: object, annotations: list[object]
) -> Argument[Any] | Option[Any] | Depends[Any]:
    for annotation in annotations:
        if isinstance(annotation, (Argument | Option | Depends)):
            return annotation

    if typing.get_origin(type_hint) is Context:
        return Depends(get_context)

    if (
        signature.kind in (signature.VAR_POSITIONAL, signature.VAR_KEYWORD)
        or signature.default is not signature.empty
    ):
        return Option()

    return Argument()



import dataclasses
import json
import pathlib
from collections.abc import Callable
from typing import Annotated, Any, Literal, ParamSpec, TypeVar

from tabb.callback import register_callback
from tabb.context import Context
from tabb.decorators import group
from tabb.parameter.types import Counter, List, Path
from tabb.types import Argument, Depends, Length, Option

T = TypeVar("T")
P = ParamSpec("P")


def load_json_config(
    *flags: str,
    default: str = "config.json",
    show_default: bool = True,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    if not flags:
        flags = ("--config",)

    localns = locals()

    def update_config(
        ctx: Context[Any],
        filename: Annotated[
            str, Option(*flags, default=default, show_default=show_default)
        ],
    ) -> None:
        try:
            with open(filename) as f:
                data = json.load(f)
        except Exception as error:
            print(f"Error loading config file: {error}")

        else:
            ctx.config.merge(data)

    def decorator(fn: Callable[P, T]) -> Callable[P, T]:
        register_callback(fn, update_config, localns=localns)
        return fn

    return decorator


@group()
@load_json_config()
def cli(foo: bool | None = None) -> None:
    print(f"{foo=}")


@cli.command
def cp(
    src: Annotated[
        list[pathlib.Path],
        Argument(type=List(Path(exists=True, readable=True), nargs="*")),
        Length(min=0, max=2),
    ],
    dest: pathlib.Path,
    verbose: Annotated[int, Option("-v", "--verbose", type=Counter())] = False,
) -> None:
    print(f"{src=} {dest=} {verbose=}")


@cli.command
def test_types(ints: list[int], strs: list[str]) -> None:
    print(f"{ints=} {strs=}")


@cli.command
def test_parameters(
    pos_only: str | None = None,
    /,
    pos_or_kw: str | None = None,
    *var_pos: int,
    kw_only: str | None = None,
    **var_kw: int,
) -> None:
    print(f"{pos_only=} {pos_or_kw=} {var_pos=} {kw_only=} {var_kw=}")


def get_home(ctx: Context[Any]) -> str | None:
    return ctx.environ.get("HOME")


def get_email(username: str | None = None, domain: str | None = None) -> str | None:
    if username is None or domain is None:
        return None
    return f"{username}@{domain}"


def higher_order_dep(
    home: Annotated[str | None, Depends(get_home)],
    email: Annotated[str | None, Depends(get_email)],
) -> str:
    return f"{home=} {email=}"


@dataclasses.dataclass
class User:
    name: str | None = None
    email: Annotated[str | None, Depends(get_email)] = None


@cli.command
def test_depends(
    ctx: Context[None],
    home: Annotated[str | None, Depends(get_home)],
    email: Annotated[str | None, Depends(get_email)],
    hod: Annotated[str, Depends(higher_order_dep)],
    user: Annotated[User, Depends(User)],
) -> None:
    print(f"{ctx=} {home=} {email=} {hod=} {user=}")


@cli.command
def hello(
    name: Annotated[str, Argument(default="World")],
    greeting: Annotated[
        str, Option(show_default=True, show_config=True, show_envvar=True)
    ] = "Hello",
    debug: Annotated[bool, Option(show_default=True)] = False,
    things: Annotated[list[str] | None, Option(help="Just some things.")] = None,
    repeat: Annotated[
        int, Option("-n", "--repeat", help="Number of times to repeat.")
    ] = 1,
) -> None:
    """Say hello."""
    for _ in range(repeat):
        print(f"{greeting} {name}!")
    print(f"Things: {things}")


@cli.command()
def train(
    # The training data
    examples_path: str,
    # Model options
    base_model: Annotated[str, Option()],
    tokenizer: Annotated[str, Option()],
    skills_path: Annotated[str, Option()],
    output_dir: str = "/defaut/output/path",
    # Training options
    seed: int = 42,
    max_seq_length: int = 128,
    logger: Literal["csv", "tensorboard"] = "csv",
    max_epocs: int = 1,
    gradient_accumulation_steps: int = 8,
    val_check_interval: int = 1000,
    train_batch_size: int = 4,
    eval_batch_size: int = 4,
    learning_rate: float = 1e-3,
    lr_start_factor: float = 1.0,
    lr_end_factor: float = 0.1,
    lr_scheduler_frequency: int = 1,
    no_grad: tuple[str, ...] = (),
) -> None:
    print(80 * "=")
    for key, value in locals().items():
        print(f"{key}={value}")


if __name__ == "__main__":
    cli(auto_config_prefix="cli", auto_envvar_prefix="CLI")



from __future__ import annotations

import dataclasses
import importlib
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Generic, ParamSpec, TypeVar, cast

from tabb.context import Context
from tabb.exceptions import UsageError
from tabb.missing import _MISSING_TYPE
from tabb.parameter import Parameter, resolve_params

T = TypeVar("T")
P = ParamSpec("P")

CALLBACK_ATTR = "__callbacks__"


def create_callback(
    callback: str | Callable[P, T],
    *,
    import_dependencies: str | None = None,
    localns: dict[str, Any] | None = None,
) -> Callback[P, T]:
    if isinstance(callback, str):
        return LazyCallback(callback, import_dependencies=import_dependencies)
    return Callback(callback, localns=localns)


def get_registered_callbacks(obj: object) -> list[Callback[..., Any]]:
    callbacks: list[Callback[..., Any]] = []
    while obj:
        callbacks.extend(getattr(obj, CALLBACK_ATTR, []))
        obj = getattr(obj, "__wrapped__", None)
    return callbacks


def register_callback(
    obj: object,
    callback: str | Callable[..., Any],
    *,
    import_dependencies: str | None = None,
    localns: dict[str, Any] | None = None,
) -> None:
    if not hasattr(obj, CALLBACK_ATTR):
        setattr(obj, CALLBACK_ATTR, [])
    callbacks: list[Callback[..., Any]] = getattr(obj, CALLBACK_ATTR)
    callbacks.append(
        create_callback(
            callback, import_dependencies=import_dependencies, localns=localns
        )
    )


@dataclasses.dataclass(frozen=True)
class Callback(Generic[P, T]):
    fn: Callable[P, T]
    params: tuple[Parameter[Any], ...]

    @property
    def name(self) -> str:
        return self.fn.__name__

    def __init__(
        self, fn: Callable[P, T], *, localns: dict[str, Any] | None = None
    ) -> None:
        object.__setattr__(self, "fn", fn)
        object.__setattr__(self, "params", tuple(resolve_params(fn, localns=localns)))

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name}>"

    def __call__(self, ctx: Context[Any]) -> T:
        for callback in get_registered_callbacks(self.fn):
            callback(ctx)

        args, kwargs = self.align_parameters(ctx)
        return self.fn(*args, **kwargs)

    def get_params(self, ctx: Context[Any]) -> list[Parameter[Any]]:
        result: list[Parameter[Any]] = []

        for callback in get_registered_callbacks(self.fn):
            result.extend(callback.get_params(ctx))

        result.extend(self.params)
        return result

    def align_parameters(self, ctx: Context[Any]) -> tuple[list[Any], dict[str, Any]]:
        stack = list(reversed(self.params))

        args = self._align_args(ctx, stack)
        kwargs = self._align_kwargs(ctx, stack)

        if not stack:
            return args, kwargs

        msg = f"Unexpected parameter: {stack[-1].name}"
        raise RuntimeError(msg)

    def _align_args(self, ctx: Context[Any], stack: list[Parameter[Any]]) -> list[Any]:
        args: list[Any] = []

        while stack:
            param = stack.pop()

            if param.kind != Parameter.POSITIONAL:
                stack.append(param)
                break

            item = ctx.values[param]

            if not item or isinstance(item.value, _MISSING_TYPE):
                return args

            args.append(item.value)

        while stack:
            param = stack.pop()

            if param.kind != Parameter.POSITIONAL_OR_KEYWORD:
                stack.append(param)
                break

            item = ctx.values[param]

            if not item or isinstance(item.value, _MISSING_TYPE):
                return args

            args.append(item.value)

        if stack and stack[-1].kind == Parameter.VAR_POSITIONAL:
            param = stack.pop()
            item = ctx.values[param]

            if not item or isinstance(item.value, _MISSING_TYPE):
                return args

            if not isinstance(item.value, Iterable):
                msg = (
                    "Expected iterable for VAR_POSITIONAL param "
                    f"{param.name}, got {item.value!r}"
                )
                raise TypeError(msg)

            args.extend(item.value)

        return args

    def _align_kwargs(
        self,
        ctx: Context[Any],
        stack: list[Parameter[Any]],
    ) -> dict[str, Any]:
        kwargs: dict[str, Any] = {}

        while stack:
            param = stack.pop()

            if param.kind not in (
                Parameter.POSITIONAL_OR_KEYWORD,
                Parameter.KEYWORD,
            ):
                stack.append(param)
                break

            item = ctx.values[param]

            if not item or isinstance(item.value, _MISSING_TYPE):
                continue

            if param.name in kwargs:
                msg = f"Got multiple values for keyword argument {param.name!r}."
                raise UsageError(msg, ctx=ctx)

            kwargs[param.name] = item.value

        if stack and stack[-1].kind == Parameter.VAR_KEYWORD:
            param = stack.pop()
            item = ctx.values[param]

            if not item or isinstance(item.value, _MISSING_TYPE):
                return kwargs

            if not isinstance(item.value, Mapping):
                msg = (
                    "Expected dict for VAR_KEYWORD param "
                    f"{param.name}, got {item.value!r}"
                )
                raise TypeError(msg)

            for key, value in item.value.items():
                if key in kwargs:
                    msg = f"Got multiple values for keyword argument {key!r}."
                    raise UsageError(msg, ctx=ctx)

                kwargs[key] = value

        return kwargs


@dataclasses.dataclass(frozen=True, slots=True)
class LazyCallback(Callback[P, T]):
    path: str
    import_dependencies: str | None = None

    def __init__(self, path: str, import_dependencies: str | None = None) -> None:
        object.__setattr__(self, "path", path)
        object.__setattr__(self, "import_dependencies", import_dependencies)

    @property
    def name(self) -> str:
        _, _, name = self.path.rpartition(":")
        return name

    def resolve(self) -> Callable[P, T]:
        module_name, _, callback_name = self.path.rpartition(":")

        try:
            module = importlib.import_module(module_name)
        except ImportError as error:
            if self.import_dependencies:
                package, _, _ = module_name.partition(".")
                message = (
                    f"{error}. Ensure `{package}[{self.import_dependencies}]` "
                    "extra is installed."
                )
                raise type(error)(message, *error.args[1:]) from error
            raise

        callback = getattr(module, callback_name)

        if not callable(callback):
            msg = f"Resolved callback {callback!r} is not callable"
            raise TypeError(msg)

        return cast(Callable[P, T], callback)

    def __getattr__(self, name: str) -> Any:
        if name in ("fn", "params"):
            Callback.__init__(self, self.resolve())
        return object.__getattribute__(self, name)



from __future__ import annotations

import abc
from collections.abc import Callable, Sequence
from typing import Any, NoReturn, TypeGuard, TypeVar

from tabb.context import Context
from tabb.exceptions import (
    BadParameter,
    MissingParameter,
    augment_usage_errors,
)
from tabb.missing import _MISSING_TYPE, MISSING
from tabb.nargs import IntNArgs, NArgs
from tabb.parameter.base import (
    Parameter,
    ParameterKind,
    ParameterSource,
    ParameterValue,
)
from tabb.parameter.types import ParameterArg, ParameterType
from tabb.utils import to_kebab, to_snake

ValueType = TypeVar("ValueType")


class ParserParameter(Parameter[ValueType]):
    def __init__(
        self,
        name: str,
        kind: ParameterKind,
        type: ParameterType[ValueType],
        *,
        config: Sequence[str] | None = None,
        default_factory: Callable[[], ValueType] | _MISSING_TYPE = MISSING,
        default: ValueType | _MISSING_TYPE = MISSING,
        envvar: Sequence[str] | None = None,
        help: str | None = None,
        hidden: bool = False,
        metavar: str | None = None,
        required: bool | None = None,
    ) -> None:
        super().__init__(name, kind)
        self.config = None if config is None else list(config)
        self.type = type
        self.default = default
        self.default_factory = default_factory
        self.envvar = None if envvar is None else list(envvar)
        self.help = help
        self.hidden = hidden
        self.metavar = metavar
        self.required = required

    @abc.abstractmethod
    def get_error_hint(self, ctx: Context[Any] | None = None) -> str:
        ...

    @property
    def nargs(self) -> NArgs:
        return self.type.nargs

    def is_required(self, ctx: Context[Any] | None = None) -> bool:
        if self.required is not None:
            return self.required

        return not (self.has_default(ctx) or self.type.has_default())

    def has_default(self, ctx: Context[Any] | None = None) -> bool:
        if self.default is not MISSING:
            return True

        if self.default_factory is not MISSING:
            return True

        return False

    def get_metavar(self, ctx: Context[Any] | None = None) -> str:
        if self.metavar is not None:
            return self.metavar

        metavar = self.type.get_metavar(ctx)
        if metavar is None:
            metavar = self.name.upper()

        return self.nargs.format_metavar(metavar)

    def get_config(self, ctx: Context[Any] | None = None) -> list[str]:
        if self.config is not None:
            return self.config

        if ctx and ctx.auto_config_prefix is not None:
            return [f"{ctx.auto_config_prefix}.{to_kebab(self.name).lower()}"]

        return []

    def get_envvar(self, ctx: Context[Any] | None = None) -> list[str]:
        if self.envvar is not None:
            return self.envvar

        if ctx and ctx.auto_envvar_prefix is not None:
            return [f"{ctx.auto_envvar_prefix}_{to_snake(self.name).upper()}"]

        return []

    def fail(
        self,
        message: str,
        ctx: Context[Any] | None = None,
    ) -> NoReturn:
        """Helper method to fail with an invalid value message."""
        raise BadParameter(message, ctx=ctx, param=self)

    def fail_missing(
        self,
        message: str | None = None,
        ctx: Context[Any] | None = None,
    ) -> NoReturn:
        """Helper method to fail with a missing value message."""
        raise MissingParameter(
            message, ctx=ctx, param=self, param_type=self.get_param_type()
        )

    def matches(
        self,
        ctx: Context[Any],
        arg: ParameterArg,
    ) -> bool:
        return True

    def process_args(
        self,
        ctx: Context[Any],
        args: list[ParameterArg],
    ) -> ParameterValue[ValueType]:
        value, source = self.get_value(ctx, args)

        if self.validate(ctx, value):
            return ParameterValue(value, source, self)

        self.fail(f"Invalid value {value!r}", ctx=ctx)

    def get_value(
        self,
        ctx: Context[Any],
        args: list[ParameterArg],
    ) -> tuple[Any | _MISSING_TYPE, ParameterSource]:
        value: Any

        with augment_usage_errors(ctx, self, parameter_errors=ValueError):
            # Only process args first if any were provided, otherwise we look
            # for a default value elsewhere first.
            if args:
                value = self.type.process_args(args)
                if not isinstance(value, _MISSING_TYPE):
                    return value, ParameterSource.COMMANDLINE

            for envar in self.get_envvar(ctx):
                if envar not in ctx.environ:
                    continue

                value = self.type.parse_envvar(ctx.environ[envar])
                return value, ParameterSource.ENVIRONMENT

            for config_path in self.get_config(ctx):
                value = ctx.config.get_path(config_path, MISSING)
                if isinstance(value, _MISSING_TYPE):
                    continue

                value = self.type.process_config(value)
                return value, ParameterSource.CONFIG

            if not isinstance(self.default, _MISSING_TYPE):
                return self.default, ParameterSource.DEFAULT

            if not isinstance(self.default_factory, _MISSING_TYPE):
                return self.default_factory(), ParameterSource.DEFAULT

            # Process are args again if we didn't find a default value
            # to see if our type will accept an empty value.
            if not args:
                value = self.type.process_args(args)

            return value, ParameterSource.COMMANDLINE

    def validate(
        self,
        ctx: Context[Any],
        value: Any | _MISSING_TYPE,
    ) -> TypeGuard[ValueType | _MISSING_TYPE]:
        if value is MISSING and self.is_required(ctx):
            self.fail_missing(ctx=ctx)

        if value is MISSING:
            return True

        with augment_usage_errors(ctx, self, parameter_errors=(TypeError, ValueError)):
            return self.type.validate(value)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name!r}>"


class ArgumentParameter(ParserParameter[ValueType]):
    @property
    def nargs(self) -> NArgs:
        nargs = super().nargs
        if self.is_required():
            return nargs
        return nargs.as_optional()

    def get_error_hint(self, ctx: Context[Any] | None = None) -> str:
        return f"'{self.get_metavar(ctx)}'"

    def get_usage_pieces(self, ctx: Context[Any] | None = None) -> list[str]:
        return [self.get_metavar(ctx)]

    def get_param_type(self) -> str:
        return "argument"

    def matches(self, ctx: Context[Any], arg: ParameterArg) -> bool:
        return super().matches(ctx, arg) and self.type.matches(arg)


class OptionParameter(ParserParameter[ValueType]):
    def __init__(
        self,
        name: str,
        kind: ParameterKind,
        type: ParameterType[ValueType],
        flags: Sequence[str],
        *,
        envvar: str | Sequence[str] | None = None,
        default_factory: Callable[[], ValueType] | _MISSING_TYPE = MISSING,
        default: ValueType | _MISSING_TYPE = MISSING,
        help: str | None = None,
        hidden: bool = False,
        metavar: str | None = None,
        required: bool | None = None,
        show_config: bool = False,
        show_default: bool | str | None = None,
        show_envvar: bool = False,
    ) -> None:
        if isinstance(type.nargs, str):
            raise ValueError("nargs for options may not be veradic.")

        self.flags, self.secondary_flags = type.parse_flags(flags)
        self.show_config = show_config
        self.show_default = show_default
        self.show_envvar = show_envvar

        super().__init__(
            name=name,
            kind=kind,
            type=type,
            envvar=envvar,
            default=default,
            default_factory=default_factory,
            help=help,
            hidden=hidden,
            metavar=metavar,
            required=required,
        )

    @property
    def nargs(self) -> IntNArgs:
        if isinstance(nargs := self.type.nargs, IntNArgs):
            return nargs
        raise ValueError("nargs for options may not be variadic.")

    def get_error_hint(self, ctx: Context[Any] | None = None) -> str:
        return " / ".join(f"'{flag}'" for flag in self.flags)

    def get_default_repr(self, ctx: Context[Any] | None = None) -> str | None:
        if isinstance(self.show_default, str):
            return f"({self.show_default})"

        if not self.show_default:
            return None

        if self.default is not None and not isinstance(self.default, _MISSING_TYPE):
            return self.type.format_value(self.default)

        if not isinstance(self.default_factory, _MISSING_TYPE):
            return "(dynamic)"

        return None

    def get_help_record(
        self, ctx: Context[Any] | None = None
    ) -> tuple[str, str] | None:
        if self.hidden:
            return None

        def _write_opts(flags: Sequence[str]) -> str:
            rv = ", ".join(sorted(flags))

            if metavar := self.get_metavar(ctx):
                rv += f" {metavar}"

            return rv

        rv = [_write_opts(self.flags)]

        if self.secondary_flags:
            rv.append(_write_opts(self.secondary_flags))

        help_text = self.help or ""
        extra = []

        if self.show_config and (config := self.get_config(ctx)):
            paths = ", ".join(config)
            extra.append(f"config path: {paths}")

        if self.show_envvar and (envvar := self.get_envvar(ctx)):
            var_str = ", ".join(envvar)
            extra.append(f"env var: {var_str}")

        if default_repr := self.get_default_repr(ctx):
            extra.append(f"default: {default_repr}")

        if self.is_required(ctx):
            extra.append("required")

        if extra:
            extra_str = "; ".join(extra)
            help_text = f"{help_text}  [{extra_str}]" if help_text else f"[{extra_str}]"

        return (" / ").join(rv), help_text

    def get_param_type(self) -> str:
        return "option"



from __future__ import annotations

import dataclasses
import re
from collections.abc import Callable, Sequence
from typing import Any, Generic, TypeVar

from tabb.missing import _MISSING_TYPE, MISSING
from tabb.parameter.types import ParameterType

ValueType = TypeVar("ValueType")


class Secret(str):
    """
    Holds a string value that should not be revealed in tracebacks etc.
    You should cast the value to `str` at the point it is required.
    """

    def __repr__(self) -> str:
        return f'{type(self).__name__}("**********")'


@dataclasses.dataclass()
class Length:
    min: int | None = None
    max: int | None = None


@dataclasses.dataclass()
class Range:
    min: int | float | None = None
    max: int | float | None = None
    max_open: bool = False
    min_open: bool = False
    clamp: bool = False


@dataclasses.dataclass()
class Choices(Generic[ValueType]):
    values: tuple[ValueType, ...]

    def __init__(self, *values: ValueType) -> None:
        self.values = values


@dataclasses.dataclass()
class Matches:
    pattern: re.Pattern[str]

    def __init__(self, pattern: str | re.Pattern[str]) -> None:
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        self.pattern = pattern


@dataclasses.dataclass()
class Greedy:
    is_greedy: bool = True


@dataclasses.dataclass()
class Validate(Generic[ValueType]):
    validator: Callable[[Any], bool]


class Argument(Generic[ValueType]):
    def __init__(
        self,
        *,
        config: str | Sequence[str] | _MISSING_TYPE = MISSING,
        default_factory: Callable[[], ValueType] | _MISSING_TYPE = MISSING,
        default: ValueType | _MISSING_TYPE = MISSING,
        envvar: str | Sequence[str] | _MISSING_TYPE = MISSING,
        help: str | _MISSING_TYPE = MISSING,
        hidden: bool | _MISSING_TYPE = MISSING,
        metavar: str | _MISSING_TYPE = MISSING,
        type: ParameterType[ValueType] | _MISSING_TYPE = MISSING,
    ) -> None:
        self.config = config
        self.default_factory = default_factory
        self.default: ValueType | _MISSING_TYPE = default
        self.envvar = envvar
        self.help = help
        self.hidden = hidden
        self.metavar = metavar
        self.type = type


class Option(Generic[ValueType]):
    def __init__(
        self,
        *flags: str,
        config: str | Sequence[str] | _MISSING_TYPE = MISSING,
        default_factory: Callable[[], Any] | _MISSING_TYPE = MISSING,
        default: Any | _MISSING_TYPE = MISSING,
        envvar: str | Sequence[str] | _MISSING_TYPE = MISSING,
        help: str | _MISSING_TYPE = MISSING,
        hidden: bool | _MISSING_TYPE = MISSING,
        metavar: str | _MISSING_TYPE = MISSING,
        show_config: bool | _MISSING_TYPE = MISSING,
        show_default: bool | _MISSING_TYPE = MISSING,
        show_envvar: bool | _MISSING_TYPE = MISSING,
        type: ParameterType[Any] | _MISSING_TYPE = MISSING,
    ) -> None:
        self.flags = flags
        self.config = config
        self.default_factory = default_factory
        self.default = default
        self.envvar = envvar
        self.help = help
        self.hidden = hidden
        self.metavar = metavar
        self.show_config = show_config
        self.show_default = show_default
        self.show_envvar = show_envvar
        self.type = type


class Depends(Generic[ValueType]):
    def __init__(
        self, dependency: Callable[..., ValueType], *, use_cache: bool = True
    ) -> None:
        self.dependency = dependency
        self.use_cache = use_cache



from __future__ import annotations

from typing import TypeGuard


class _MISSING_TYPE:  # noqa: N801
    pass


MISSING = _MISSING_TYPE()


def is_missing(value: object) -> TypeGuard[_MISSING_TYPE]:
    return value is MISSING
